import logging
import os
import re

from abc import ABC, abstractmethod
from typing import Dict, Optional
from urllib.parse import urlparse

from .._model import AccountConfig, ServerConfig, TransportEncryption

email_match_re = re.compile(r'[^@]+@[^@]+\.[^@]+')


class BaseMailPlugin(ABC):  # pylint: disable=too-few-public-methods
    """
    Base class for mail plugins.
    """

    def __init__(
        self,
        server: str,
        account: AccountConfig,
        timeout: float,
        keyfile: Optional[str] = None,
        certfile: Optional[str] = None,
        domain: Optional[str] = None,
        **_,
    ):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.account = account
        self.server = self._get_server_config(
            server=server,
            timeout=timeout,
            keyfile=keyfile,
            certfile=certfile,
            domain=domain,
        )

    @staticmethod
    def _get_path(path: str) -> str:
        return os.path.abspath(os.path.expanduser(path))

    @classmethod
    def _get_server_config(
        cls,
        server: str,
        timeout: float,
        keyfile: Optional[str],
        certfile: Optional[str],
        domain: Optional[str],
    ) -> ServerConfig:
        url = urlparse(server)
        assert url.hostname, f'No hostname specified: "{server}"'

        ssl = TransportEncryption.by_url_scheme(url.scheme)
        port = url.port or cls.default_ports().get(ssl)
        assert port, f'No port specified and no default available: "{server}"'

        if keyfile:
            keyfile = cls._get_path(keyfile)
        if certfile:
            certfile = cls._get_path(certfile)

        return ServerConfig(
            server=url.hostname,
            port=port,
            encryption=ssl,
            timeout=timeout,
            keyfile=keyfile,
            certfile=certfile,
            domain=domain,
        )

    @property
    def from_string(self):
        """
        :return: The from string for the account.
        """
        return self.account.display_name or self.account.username

    @classmethod
    @abstractmethod
    def _matches_url_scheme(cls, scheme: str) -> bool:
        raise NotImplementedError()

    @classmethod
    def can_handle(cls, url: str) -> bool:
        """
        Check whether the plugin can handle the specified URL.
        """
        return cls._matches_url_scheme(urlparse(url).scheme)

    @classmethod
    @abstractmethod
    def default_ports(cls) -> Dict[TransportEncryption, int]:
        """
        :return: A mapping of transport encryption to default port.
        """
        raise NotImplementedError()

    @classmethod
    def build(
        cls,
        server: str,
        account: AccountConfig,
        timeout: float,
        keyfile: Optional[str] = None,
        certfile: Optional[str] = None,
    ) -> 'BaseMailPlugin':
        from ._utils import mail_plugins

        url_parsed = urlparse(server)
        assert url_parsed.hostname, f'No hostname specified: "{server}"'

        mail_cls = next(
            (plugin for plugin in mail_plugins if plugin.can_handle(server)),
            None,
        )

        assert mail_cls, f'No mail plugin found for URL: "{server}"'
        return mail_cls(
            server=server,
            account=account,
            timeout=timeout,
            keyfile=keyfile,
            certfile=certfile,
        )


# vim:sw=4:ts=4:et:



from abc import ABC

from platypush.message.event import Event


class JoystickEvent(Event, ABC):
    """
    Base joystick event class.
    """

    def __init__(self, device: dict, *args, **kwargs):
        """
        :param device: Joystick device info as a dictionary:

            .. schema:: joystick.JoystickDeviceSchema
        """
        super().__init__(*args, device=device, **kwargs)


class JoystickConnectedEvent(JoystickEvent):
    """
    Event triggered upon joystick connection.
    """


class JoystickDisconnectedEvent(JoystickEvent):
    """
    Event triggered upon joystick disconnection.
    """


class JoystickStateEvent(JoystickEvent):
    """
    Base joystick state event class.
    """

    def __init__(self, *args, state: dict, **kwargs):
        """
        :param state: Joystick state as a dictionary:

            .. schema:: joystick.JoystickStateSchema
        """
        super().__init__(*args, state=state, **kwargs)


# vim:sw=4:ts=4:et:



from typing import Union

from platypush.entities import Entity
from platypush.message.event import Event


class EntityEvent(Event):
    def __init__(self, entity: Union[Entity, dict], *args, **kwargs):
        if isinstance(entity, Entity):
            entity = entity.to_dict()
        super().__init__(entity=entity, *args, **kwargs)


class EntityUpdateEvent(EntityEvent):
    """
    This event is triggered whenever an entity of any type (a switch, a light,
    a sensor, a media player etc.) updates its state.
    """


class EntityDeleteEvent(EntityEvent):
    """
    This event is triggered whenever an entity is deleted.
    """


# vim:sw=4:ts=4:et:



import importlib
import inspect
import os

from threading import RLock
from typing import Set, Type

import pkgutil

from ._base import BaseMailPlugin


def _scan_mail_plugins() -> Set[Type[BaseMailPlugin]]:
    from platypush.plugins import mail

    # Recursively scan for class inside the `mail` module that inherit from
    # BaseMailPlugin
    base_file = inspect.getfile(mail)
    plugins = set()

    for _, name, _ in pkgutil.walk_packages(
        [os.path.dirname(base_file)], prefix=f'{mail.__name__}.'
    ):
        module = importlib.import_module(name)
        for _, cls in inspect.getmembers(module, inspect.isclass):
            if not inspect.isabstract(cls) and issubclass(cls, BaseMailPlugin):
                plugins.add(cls)

    return plugins


_mail_plugins_lock = RLock()
mail_plugins = set()

with _mail_plugins_lock:
    if not mail_plugins:
        mail_plugins = _scan_mail_plugins()


# vim:sw=4:ts=4:et:



import os
from typing import Optional

from platypush.message.event import Event


class InotifyEvent(Event):
    """
    Generic super-class for inotify events.
    """
    def __init__(self, path: str, resource: Optional[str] = None, resource_type: Optional[str] = None,
                 *args, **kwargs):
        """
        :param path: Monitored path.
        :param resource: File/resource name.
        :param resource_type: INotify type of the resource, if available.
        """
        kwargs['full_path'] = os.path.join(path, resource) if resource else path
        super().__init__(*args, path=path, resource=resource,
                         resource_type=self._resource_type_code_to_name(resource_type), **kwargs)

    @staticmethod
    def _resource_type_code_to_name(resource_type: Optional[str] = None) -> Optional[str]:
        if resource_type == 'IN_ISDIR':
            return 'directory'

        return resource_type or 'file'


class InotifyOpenEvent(InotifyEvent):
    """
    Event triggered when a monitored resource is opened.
    """


class InotifyCloseEvent(InotifyEvent):
    """
    Event triggered when a monitored resource is closed.
    """


class InotifyAccessEvent(InotifyEvent):
    """
    Event triggered when a monitored resource is accessed.
    """


class InotifyCreateEvent(InotifyEvent):
    """
    Event triggered when a monitored resource is created.
    """


class InotifyDeleteEvent(InotifyEvent):
    """
    Event triggered when a monitored resource is deleted.
    """


class InotifyModifyEvent(InotifyEvent):
    """
    Event triggered when a monitored resource is modified.
    """


class InotifyMovedEvent(InotifyEvent):
    """
    Event triggered when a resource in a monitored path is moved.
    """
    def __init__(self, path: str, old: Optional[str] = None, new: Optional[str] = None, *args, **kwargs):
        """
        :param path: Monitored path.
        :param old: Old name.
        :param new: New name.
        """
        super().__init__(path=path, old=old, new=new, *args, **kwargs)


class InotifyPermissionsChangeEvent(InotifyEvent):
    """
    Event triggered when the permissions on a monitored resource are changed.
    """
    def __init__(self, path: str, umask: int, resource: Optional[str] = None, *args, **kwargs):
        """
        :param path: Monitored path.
        :param umask: New umask.
        :param resource: File/resource name.
        """
        super().__init__(path=path, resource=resource, umask=umask, *args, **kwargs)


# vim:sw=4:ts=4:et:



from sqlalchemy import Boolean, Column, Float, ForeignKey, Integer, JSON, String

from platypush.common.db import is_defined

from . import Entity
from .devices import Device
from .sensors import NumericSensor, PercentSensor
from .temperature import TemperatureSensor


if not is_defined('cpu'):

    class Cpu(Entity):
        """
        ``CPU`` ORM (container) model.
        """

        __tablename__ = 'cpu'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        percent = Column(Float)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('cpu_info'):

    class CpuInfo(Entity):
        """
        ``CpuInfo`` ORM model.
        """

        __tablename__ = 'cpu_info'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        architecture = Column(String)
        bits = Column(Integer)
        cores = Column(Integer)
        vendor = Column(String)
        brand = Column(String)
        frequency_advertised = Column(Integer)
        frequency_actual = Column(Integer)
        flags = Column(JSON)
        l1_instruction_cache_size = Column(Integer)
        l1_data_cache_size = Column(Integer)
        l2_cache_size = Column(Integer)
        l3_cache_size = Column(Integer)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('cpu_times'):

    class CpuTimes(Entity):
        """
        ``CpuTimes`` ORM (container) model.
        """

        __tablename__ = 'cpu_times'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('cpu_stats'):

    class CpuStats(Entity):
        """
        ``CpuStats`` ORM (container) model.
        """

        __tablename__ = 'cpu_stats'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('memory_stats'):

    class MemoryStats(Entity):
        """
        ``MemoryStats`` ORM model.
        """

        __tablename__ = 'memory_stats'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        total = Column(Integer)
        available = Column(Integer)
        used = Column(Integer)
        free = Column(Integer)
        active = Column(Integer)
        inactive = Column(Integer)
        buffers = Column(Integer)
        cached = Column(Integer)
        shared = Column(Integer)
        percent = Column(Float)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('swap_stats'):

    class SwapStats(Entity):
        """
        ``SwapStats`` ORM model.
        """

        __tablename__ = 'swap_stats'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        total = Column(Integer)
        used = Column(Integer)
        free = Column(Integer)
        percent = Column(Float)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('disk'):

    class Disk(Entity):
        """
        ``Disk`` ORM model.
        """

        __tablename__ = 'disk'

        id = Column(
            Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True
        )

        mountpoint = Column(String)
        fstype = Column(String)
        opts = Column(String)
        total = Column(Integer)
        used = Column(Integer)
        free = Column(Integer)
        percent = Column(Float)
        read_count = Column(Integer)
        write_count = Column(Integer)
        read_bytes = Column(Integer)
        write_bytes = Column(Integer)
        read_time = Column(Float)
        write_time = Column(Float)
        busy_time = Column(Float)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('network_interface'):

    class NetworkInterface(Device):
        """
        ``NetworkInterface`` ORM model.
        """

        __tablename__ = 'network_interface'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )

        bytes_sent = Column(Integer)
        bytes_recv = Column(Integer)
        packets_sent = Column(Integer)
        packets_recv = Column(Integer)
        errors_in = Column(Integer)
        errors_out = Column(Integer)
        drop_in = Column(Integer)
        drop_out = Column(Integer)
        addresses = Column(JSON)
        speed = Column(Integer)
        mtu = Column(Integer)
        duplex = Column(String)
        flags = Column(JSON)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('system_temperature'):

    class SystemTemperature(TemperatureSensor):
        """
        Extends the ``TemperatureSensor``.
        """

        __tablename__ = 'system_temperature'

        id = Column(
            Integer,
            ForeignKey(TemperatureSensor.id, ondelete='CASCADE'),
            primary_key=True,
        )

        high = Column(Float)
        critical = Column(Float)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('system_fan'):

    class SystemFan(NumericSensor):
        """
        ``SystemFan`` ORM model.
        """

        __tablename__ = 'system_fan'

        id = Column(
            Integer,
            ForeignKey(NumericSensor.id, ondelete='CASCADE'),
            primary_key=True,
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('system_battery'):

    class SystemBattery(PercentSensor):
        """
        ``SystemBattery`` ORM model.
        """

        __tablename__ = 'system_battery'

        id = Column(
            Integer,
            ForeignKey(PercentSensor.id, ondelete='CASCADE'),
            primary_key=True,
        )

        seconds_left = Column(Float)
        power_plugged = Column(Boolean)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from typing import Optional

from platypush.message.event import Event


class AlarmEvent(Event):
    """
    Base class for alarm events.
    """

    def __init__(self, *args, name: Optional[str] = None, **kwargs):
        super().__init__(*args, name=name, **kwargs)


class AlarmEnabledEvent(AlarmEvent):
    """
    Triggered when an alarm is enabled.
    """


class AlarmDisabledEvent(AlarmEvent):
    """
    Triggered when an alarm is disabled.
    """


class AlarmStartedEvent(AlarmEvent):
    """
    Triggered when an alarm starts.
    """


class AlarmEndedEvent(AlarmEvent):
    """
    Triggered when an alarm stops.
    """


class AlarmDismissedEvent(AlarmEndedEvent):
    """
    Triggered when an alarm is dismissed.
    """


class AlarmSnoozedEvent(AlarmEvent):
    """
    Triggered when an alarm is snoozed.
    """


class AlarmTimeoutEvent(AlarmEndedEvent):
    """
    Triggered when an alarm times out.
    """


# vim:sw=4:ts=4:et:



from sqlalchemy import Column, Integer, ForeignKey

from platypush.common.db import is_defined

from .three_axis import ThreeAxisSensor


if not is_defined('magnetometer'):

    class Magnetometer(ThreeAxisSensor):
        """
        An entity that represents the value of a magnetometer.
        """

        __tablename__ = 'magnetometer'

        id = Column(
            Integer,
            ForeignKey(ThreeAxisSensor.id, ondelete='CASCADE'),
            primary_key=True,
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



import aioxmpp
import aioxmpp.dispatcher

from platypush.message.event.xmpp import (
    XmppMessageReceivedEvent,
)

from ._base import XmppBaseHandler


# pylint: disable=too-many-ancestors,too-few-public-methods
class XmppMessageHandler(XmppBaseHandler):
    """
    Handler for XMPP message events.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dispatcher = self._client.summon(
            aioxmpp.dispatcher.SimpleMessageDispatcher
        )
        self.dispatcher.register_callback(
            aioxmpp.MessageType.CHAT,
            None,  # from filter
            self._on_msg_received,
        )

    def _on_msg_received(self, msg, *_, **__):
        if not msg.body:
            return

        if msg.error:
            self.logger.warning('Error on message from %s: %s', msg.from_, msg.error)

        body = msg.body.lookup([aioxmpp.structs.LanguageRange.fromstr('*')])
        self._post_user_event(
            XmppMessageReceivedEvent, user_id=msg.from_, body=body.rstrip()
        )



from abc import ABC

from ._events import XmppEventMixin
from ._state import XmppStateMixin


# pylint: disable=too-few-public-methods
class XmppEventStateMixin(XmppEventMixin, XmppStateMixin, ABC):
    """
    A mixin that encapsulates the state of the XMPP clients and it provides the
    features to handle events.
    """

    def _post_event(self, *args, **kwargs):
        if self._state_serializer:
            self._state_serializer.enqueue(self._state)

        return super()._post_event(*args, **kwargs)



from abc import ABC, abstractmethod
from logging import getLogger
from threading import Thread

from tornado.ioloop import IOLoop
from tornado.websocket import WebSocketHandler

from platypush.backend.http.app.utils.auth import AuthStatus, get_auth_status

from ..mixins import MessageType, PubSubMixin

logger = getLogger(__name__)


class WSRoute(WebSocketHandler, Thread, PubSubMixin, ABC):
    """
    Base class for Tornado websocket endpoints.
    """

    def __init__(self, *args, **kwargs):
        WebSocketHandler.__init__(self, *args)
        PubSubMixin.__init__(self, **kwargs)
        Thread.__init__(self)
        self._io_loop = IOLoop.current()

    def open(self, *_, **__):
        auth_status = get_auth_status(self.request)
        if auth_status != AuthStatus.OK:
            self.close(code=1008, reason=auth_status.value.message)  # Policy Violation
            return

        logger.info(
            'Client %s connected to %s', self.request.remote_ip, self.request.path
        )
        self.name = f'ws:{self.app_name()}@{self.request.remote_ip}'
        self.start()

    def data_received(self, *_, **__):
        pass

    def on_message(self, message):
        return message

    @classmethod
    @abstractmethod
    def app_name(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def path(cls) -> str:
        return f'/ws/{cls.app_name()}'

    @property
    def auth_required(self):
        return True

    def send(self, msg: MessageType) -> None:
        self._io_loop.asyncio_loop.call_soon_threadsafe(  # type: ignore
            self.write_message, self._serialize(msg)
        )

    def run(self) -> None:
        super().run()
        self.subscribe(*self._subscriptions)

    def on_close(self):
        super().on_close()
        for channel in self._subscriptions.copy():
            self.unsubscribe(channel)

        if self._pubsub:
            self._pubsub.close()

        logger.info(
            'Client %s disconnected from %s, reason=%s, message=%s',
            self.request.remote_ip,
            self.request.path,
            self.close_code,
            self.close_reason,
        )



from platypush.plugins import Plugin, action


class ChatPlugin(Plugin):
    """
    Base class for chat plugins.
    """

    @action
    def send_message(self, *_, **__):
        raise NotImplementedError()



from platypush.message.event import Event


class TidalEvent(Event):
    """Base class for Tidal events"""


class TidalPlaylistUpdatedEvent(TidalEvent):
    """
    Event fired when a Tidal playlist is updated.
    """

    def __init__(self, playlist_id: str, *args, **kwargs):
        super().__init__(*args, playlist_id=playlist_id, **kwargs)



import hashlib
import requests

from typing import Any, Dict, Optional, Union, List

from platypush.schemas.pihole import PiholeStatusSchema
from platypush.plugins import Plugin, action


class PiholePlugin(Plugin):
    """
    Plugin for interacting with a `Pi-Hole <https://pi-hole.net>`_ DNS server for advertisement and content blocking.
    """

    def __init__(
        self,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: bool = False,
        verify_ssl: bool = True,
        timeout: int = 10,
        **kwargs,
    ):
        """
        :param server: Default Pi-hole server IP address.
        :param password: Password for the default Pi-hole server.
        :param api_key: Alternatively to the password, you can also provide the server ``api_key``, as retrieved from
            http://pi-hole-server/admin/scripts/pi-hole/php/api_token.php
        :param ssl: Set to true if the host uses HTTPS (default: False).
        :param verify_ssl: Set to False to disable SSL certificate check.
        :param timeout: Default timeout for the HTTP requests.
        """
        super().__init__(**kwargs)
        self.server = server
        self.password = password
        self.api_key = api_key
        self.ssl = ssl
        self.verify_ssl = verify_ssl
        self.timeout = timeout

    @staticmethod
    def _get_token(
        password: Optional[str] = None, api_key: Optional[str] = None
    ) -> str:
        if not password:
            return api_key or ''
        return hashlib.sha256(
            hashlib.sha256(str(password).encode()).hexdigest().encode()
        ).hexdigest()  # lgtm [py/weak-sensitive-data-hashing]

    def _get_url(
        self,
        name: str,
        server: Optional[str] = None,
        password: Optional[str] = None,
        ssl: Optional[bool] = None,
        api_key: Optional[str] = None,
        **kwargs,
    ) -> str:
        if not server:
            server = self.server
            password = password or self.password
            api_key = api_key or self.api_key
            ssl = ssl if ssl is not None else self.ssl

        args = '&'.join(
            [f'{key}={value}' for key, value in kwargs.items() if value is not None]
        )

        if args:
            args = '&' + args

        token = self._get_token(password=password, api_key=api_key)
        if token:
            token = '&auth=' + token

        return f'http{"s" if ssl else ""}://{server}/admin/api.php?{name}{token}{args}'

    @staticmethod
    def _normalize_number(n: Union[str, int]):
        return int(str(n).replace(',', ''))

    @action
    def status(
        self,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ) -> dict:
        """
        Get the status and statistics of a running Pi-hole server.

        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        :return: .. schema:: pihole.PiholeStatusSchema
        """
        status = requests.get(
            self._get_url(
                'summary', server=server, password=password, api_key=api_key, ssl=ssl
            ),
            verify=self.verify_ssl,
            timeout=self.timeout,
        ).json()

        version = requests.get(
            self._get_url(
                'versions', server=server, password=password, api_key=api_key, ssl=ssl
            ),
            verify=self.verify_ssl,
            timeout=self.timeout,
        ).json()

        return dict(
            PiholeStatusSchema().dump(
                {
                    'server': server or self.server,
                    'status': status.get('status'),
                    'ads_percentage': float(status.get('ads_percentage_today')),
                    'blocked': self._normalize_number(status.get('ads_blocked_today')),
                    'cached': self._normalize_number(status.get('queries_cached')),
                    'domain_count': self._normalize_number(
                        status.get('domains_being_blocked')
                    ),
                    'forwarded': self._normalize_number(
                        status.get('queries_forwarded')
                    ),
                    'queries': self._normalize_number(status.get('dns_queries_today')),
                    'total_clients': self._normalize_number(
                        status.get('clients_ever_seen')
                    ),
                    'total_queries': self._normalize_number(
                        status.get('dns_queries_all_types')
                    ),
                    'unique_clients': self._normalize_number(
                        status.get('unique_clients')
                    ),
                    'unique_domains': self._normalize_number(
                        status.get('unique_domains')
                    ),
                    'version': version.get('core_current'),
                }
            )
        )

    @action
    def enable(
        self,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ):
        """
        Enable a Pi-hole server.

        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        response = requests.get(
            self._get_url(
                'enable', server=server, password=password, api_key=api_key, ssl=ssl
            ),
            verify=self.verify_ssl,
            timeout=self.timeout,
        )

        try:
            status = (response.json() or {}).get('status')
            assert status == 'enabled', 'Wrong credentials'
        except Exception as e:
            raise AssertionError(f'Could not enable the server: {response.text or e}')

        return response.json()

    @action
    def disable(
        self,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        seconds: Optional[int] = None,
        ssl: Optional[bool] = None,
    ):
        """
        Disable a Pi-hole server.

        :param seconds: How long the server will be disabled in seconds (default: None, indefinitely).
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        if seconds:
            response = requests.get(
                self._get_url(
                    '',
                    server=server,
                    password=password,
                    api_key=api_key,
                    ssl=ssl,
                    disable=seconds,
                ),
                verify=self.verify_ssl,
                timeout=self.timeout,
            )
        else:
            response = requests.get(
                self._get_url(
                    'disable',
                    server=server,
                    password=password,
                    api_key=api_key,
                    ssl=ssl,
                ),
                verify=self.verify_ssl,
                timeout=self.timeout,
            )

        try:
            status = (response.json() or {}).get('status')
            assert status == 'disabled', 'Wrong credentials'
        except Exception as e:
            raise AssertionError(f'Could not disable the server: {response.text or e}')

        return response.json()

    def _list_manage(
        self,
        domain: str,
        list_name: str,
        endpoint: str,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ):
        data: Dict[str, Any] = {'list': list_name, 'domain': domain}

        if password or self.password:
            data['pw'] = password or self.password
        elif api_key or self.api_key:
            data['auth'] = api_key or self.api_key

        base_url = (
            f"http{'s' if ssl or (ssl is None and self.ssl) else ''}://"
            f"{server or self.server}/admin/scripts/pi-hole/php/{endpoint}.php"
        )

        with requests.session() as s:
            s.get(base_url, verify=self.verify_ssl)
            response = requests.post(
                base_url, data=data, verify=self.verify_ssl, timeout=self.timeout
            ).text.strip()

        return {'response': response}

    def _list_get(
        self, list_name: str, server: Optional[str] = None, ssl: Optional[bool] = None
    ) -> List[str]:
        response = requests.get(
            f"http{'s' if ssl or (ssl is None and self.ssl) else ''}"
            f"://{server or self.server}/admin/scripts/pi-hole/php/"
            f"get.php?list={list_name}",
            verify=self.verify_ssl,
            timeout=10,
        ).json()

        ret = set()
        for ll in response:
            ret.update(ll)
        return list(ret)

    @action
    def get_blacklist(
        self, server: Optional[str] = None, ssl: Optional[bool] = None
    ) -> List[str]:
        """
        Get the content of the blacklist.

        :param server: Server IP address (default: default configured ``server`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_get(list_name='black', server=server, ssl=ssl)

    @action
    def get_whitelist(
        self, server: Optional[str] = None, ssl: Optional[bool] = None
    ) -> List[str]:
        """
        Get the content of the whitelist.

        :param server: Server IP address (default: default configured ``server`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_get(list_name='white', server=server, ssl=ssl)

    @action
    def get_list(
        self, list_name: str, server: Optional[str] = None, ssl: Optional[bool] = None
    ) -> List[str]:
        """
        Get the content of a list stored on the server.

        :param list_name: List name
        :param server: Server IP address (default: default configured ``server`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_get(list_name=list_name, server=server, ssl=ssl)

    @action
    def blacklist_add(
        self,
        domain: str,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ):
        """
        Add a domain to the blacklist.

        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(
            domain=domain,
            list_name='black',
            endpoint='add',
            server=server,
            password=password,
            api_key=api_key,
            ssl=ssl,
        )

    @action
    def blacklist_remove(
        self,
        domain: str,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ):
        """
        Remove a domain from the blacklist.

        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(
            domain=domain,
            list_name='black',
            endpoint='sub',
            server=server,
            password=password,
            api_key=api_key,
            ssl=ssl,
        )

    @action
    def whitelist_add(
        self,
        domain: str,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ):
        """
        Add a domain to the whitelist.

        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(
            domain=domain,
            list_name='white',
            endpoint='add',
            server=server,
            password=password,
            api_key=api_key,
            ssl=ssl,
        )

    @action
    def whitelist_remove(
        self,
        domain: str,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ):
        """
        Remove a domain from the whitelist.

        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(
            domain=domain,
            list_name='white',
            endpoint='sub',
            server=server,
            password=password,
            api_key=api_key,
            ssl=ssl,
        )

    @action
    def list_add(
        self,
        list_name: str,
        domain: str,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ):
        """
        Add a domain to a custom list stored on the server.

        :param list_name: List name
        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(
            domain=domain,
            list_name=list_name,
            endpoint='add',
            server=server,
            password=password,
            api_key=api_key,
            ssl=ssl,
        )

    @action
    def list_remove(
        self,
        list_name: str,
        domain: str,
        server: Optional[str] = None,
        password: Optional[str] = None,
        api_key: Optional[str] = None,
        ssl: Optional[bool] = None,
    ):
        """
        Remove a domain from a custom list stored on the server.

        :param list_name: List name
        :param domain: Domain name.
        :param server: Server IP address (default: default configured ``server`` value).
        :param password: Server password (default: default configured ``password`` value).
        :param api_key: Server API key (default: default configured ``api_key`` value).
        :param ssl: Set to True if the server uses SSL (default: False).
        """
        return self._list_manage(
            domain=domain,
            list_name=list_name,
            endpoint='sub',
            server=server,
            password=password,
            api_key=api_key,
            ssl=ssl,
        )


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod

from .switches import SwitchEntityManager


class LightEntityManager(SwitchEntityManager, ABC):
    """
    Base class for integrations that support light/bulbs entities.
    """

    @abstractmethod
    def set_lights(self, *args, lights=None, **kwargs):
        """
        Set a set of properties on a set of lights.

        :param light: List of lights to set. Each item can represent a light
            name or ID.
        :param kwargs: key-value list of the parameters to set.
        """
        raise NotImplementedError()



from typing import Any, Dict, List

from platypush.entities.acceleration import Accelerometer
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


# pylint: disable=too-many-ancestors
class SensorLis3dhPlugin(SensorPlugin):
    """
    Plugin to interact with an `Adafruit LIS3DH accelerometer
    <https://www.adafruit.com/product/2809>`_ and get X,Y,Z measurement. Tested
    with a Raspberry Pi over I2C connection.
    """

    def __init__(self, g=4, precision=None, poll_interval=1, **kwargs):
        """
        Only LIS3DH in I2C mode is currently supported: https://learn.adafruit.com/assets/59080.

        :param g: Accelerometer range as a multiple of G - can be 2G, 4G, 8G or 16G
        :type g: int

        :param precision: If set, the position values will be rounded to the specified number of decimal digits
            (default: no rounding)
        :type precision: int
        """

        from .lib.LIS3DH import LIS3DH

        super().__init__(poll_interval=poll_interval, **kwargs)

        if g == 2:
            self.g = LIS3DH.RANGE_2G
        elif g == 4:
            self.g = LIS3DH.RANGE_4G
        elif g == 8:
            self.g = LIS3DH.RANGE_8G
        elif g == 16:
            self.g = LIS3DH.RANGE_16G
        else:
            raise RuntimeError(f'Invalid G range: {g}')

        self.precision = precision
        self.sensor = LIS3DH()
        self.sensor.setRange(self.g)

    @action
    def get_measurement(self, *_, **__):
        """
        :returns: The sensor's current position as a dictionary with the three components (x,y,z) in degrees, each
            between -90 and 90
        """

        values = [
            (pos * 100 if self.precision is None else round(pos * 100, self.precision))
            for pos in (self.sensor.getX(), self.sensor.getY(), self.sensor.getZ())
        ]

        return {
            'name': 'position',
            'value': {'x': values[0], 'y': values[1], 'z': values[2]},
        }

    def transform_entities(self, entities: Dict[str, Any]) -> List[Accelerometer]:
        return Accelerometer(
            id='lis3dh',
            name='LIS3DH Accelerometer',
            value=entities['value'],
        )


# vim:sw=4:ts=4:et:



import logging
import re
import subprocess
import sys
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from typing import Collection, Dict, Iterable, Optional, List

from platypush.entities import EntityManager
from platypush.entities.ping import PingHost
from platypush.message.event.ping import HostUpEvent, HostDownEvent, PingResponseEvent
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.ping import PingResponseSchema

PING_MATCHER = re.compile(
    r"(?P<min>\d+.\d+)/(?P<avg>\d+.\d+)/(?P<max>\d+.\d+)/(?P<mdev>\d+.\d+)"
)

PING_MATCHER_BUSYBOX = re.compile(r"(?P<min>\d+.\d+)/(?P<avg>\d+.\d+)/(?P<max>\d+.\d+)")

WIN32_PING_MATCHER = re.compile(r"(?P<min>\d+)ms.+(?P<max>\d+)ms.+(?P<avg>\d+)ms")


def ping(host: str, ping_cmd: List[str], logger: logging.Logger) -> dict:
    out = None
    pinger = None
    err_response = dict(
        PingResponseSchema().dump(
            {
                "host": host,
                "success": False,
            }
        )
    )

    try:
        with subprocess.Popen(
            ping_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as pinger:
            out = pinger.communicate()
            if sys.platform == "win32":
                match = WIN32_PING_MATCHER.search(str(out).rsplit("\n", maxsplit=1)[-1])
                min_val, avg_val, max_val = match.groups()
                mdev_val = None
            elif "max/" not in str(out):
                match = PING_MATCHER_BUSYBOX.search(
                    str(out).rsplit("\n", maxsplit=1)[-1]
                )
                assert match is not None, out
                min_val, avg_val, max_val = match.groups()
                mdev_val = None
            else:
                match = PING_MATCHER.search(str(out).rsplit("\n", maxsplit=1)[-1])
                assert match is not None, out
                min_val, avg_val, max_val, mdev_val = match.groups()

            return dict(
                PingResponseSchema().dump(
                    {
                        "host": host,
                        "success": True,
                        "min": float(min_val),
                        "max": float(max_val),
                        "avg": float(avg_val),
                        "mdev": float(mdev_val) if mdev_val is not None else None,
                    }
                )
            )
    except Exception as e:
        err = (
            '\n'.join(line.decode().strip() for line in out)
            if isinstance(out, (tuple, list))
            else str(e)
        )

        logger.warning("Error while pinging host %s: %s", host, err)
        if pinger and pinger.poll() is None:
            pinger.kill()
            pinger.wait()

        return err_response


class PingPlugin(RunnablePlugin, EntityManager):
    """
    This integration allows you to:

        - Programmatic ping a remote host.
        - Monitor the status of a remote host.
    """

    def __init__(
        self,
        executable: str = 'ping',
        count: int = 1,
        timeout: float = 5.0,
        hosts: Optional[List[str]] = None,
        poll_interval: float = 20.0,
        **kwargs,
    ):
        """
        :param executable: Path to the ``ping`` executable. Default: the first ``ping`` executable found in PATH.
        :param count: Default number of packets that should be sent (default: 1).
        :param timeout: Default timeout before failing a ping request (default: 5 seconds).
        :param hosts: List of hosts to monitor. If not specified then no hosts will be monitored.
        :param poll_interval: How often the hosts should be monitored (default: 10 seconds).
        """

        super().__init__(poll_interval=poll_interval, **kwargs)
        self.executable = executable
        self.count = count
        self.timeout = timeout
        self.hosts: Dict[str, Optional[dict]] = {h: None for h in (hosts or [])}

    def _get_ping_cmd(self, host: str, count: int, timeout: float) -> List[str]:
        if sys.platform == 'win32':
            return [  # noqa
                self.executable,
                '-n',
                str(count or self.count),
                '-w',
                str((timeout or self.timeout) * 1000),
                host,
            ]

        return [
            self.executable,
            '-n',
            '-q',
            '-c',
            str(count or self.count),
            '-W',
            str(timeout or self.timeout),
            host,
        ]

    @action
    def ping(
        self, host: str, count: Optional[int] = None, timeout: Optional[float] = None
    ) -> dict:
        """
        Ping a remote host.

        :param host: Remote host IP or name
        :param count: Overrides the configured number of packets that should be
            sent (default: 1).
        :param timeout: Overrides the configured timeout before failing a ping
            request (default: 5 seconds).
        :return: .. schema:: ping.PingResponseSchema
        """
        return self._ping(host=host, count=count, timeout=timeout)

    def _ping(
        self, host: str, count: Optional[int] = None, timeout: Optional[float] = None
    ) -> dict:
        return ping(
            host=host,
            ping_cmd=self._get_ping_cmd(
                host=host, count=count or self.count, timeout=timeout or self.timeout
            ),
            logger=self.logger,
        )

    def _process_ping_result(self, result: dict):
        host = result.get("host")
        if host is None:
            return

        success = result.get("success")
        if success is None:
            return

        prev_result = self.hosts.get(host)
        if success:
            if not (prev_result and prev_result.get("success")):
                self._bus.post(HostUpEvent(host=host))

            self._bus.post(PingResponseEvent(**result))
            self.hosts[host] = result
        else:
            if not prev_result or prev_result.get("success"):
                self._bus.post(HostDownEvent(host=host))
            self.hosts[host] = result

    @action
    def status(self) -> Collection[PingHost]:
        """
        Get the status of the monitored hosts.
        :return: Dictionary of monitored hosts and their status.
        """
        return self.publish_entities()

    def publish_entities(self, *_, **__) -> Collection[PingHost]:
        return super().publish_entities(self.hosts.values())

    def transform_entities(
        self, entities: Collection[Optional[dict]], **_
    ) -> Iterable[PingHost]:
        return super().transform_entities(
            [
                PingHost(
                    id=status.get("host"),
                    name=status.get("host"),
                    reachable=status.get("success"),
                    min=status.get("min"),
                    max=status.get("max"),
                    avg=status.get("avg"),
                )
                for status in entities
                if status
            ]
        )

    def main(self):
        # Don't start the thread if no monitored hosts are configured
        if not self.hosts:
            self.wait_stop()
            return

        while not self.should_stop():
            try:
                with ProcessPoolExecutor() as executor:
                    for result in executor.map(
                        ping,
                        self.hosts.keys(),
                        [
                            self._get_ping_cmd(h, self.count, self.timeout)
                            for h in self.hosts.keys()
                        ],
                        [self.logger] * len(self.hosts),
                    ):
                        self._process_ping_result(result)
            except (KeyboardInterrupt, BrokenProcessPool):
                break
            except Exception as e:
                self.logger.warning("Error while pinging hosts: %s", e)
                self.logger.exception(e)
            finally:
                self.publish_entities()
                self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



from ._base import BaseBluetoothClass, ClassProperty


class MajorServiceClass(BaseBluetoothClass):
    """
    Models Bluetooth major service classes - see
    https://btprodspecificationrefs.blob.core.windows.net/assigned-numbers/Assigned%20Number%20Types/Assigned%20Numbers.pdf,
    Section 2.8.1
    """

    LE_AUDIO = ClassProperty('Low-energy Audio', 1 << 14, 14)
    POSITIONING = ClassProperty('Positioning', 1 << 16, 16)
    NETWORKING = ClassProperty('Networking', 1 << 17, 17)
    RENDERING = ClassProperty('Rendering', 1 << 18, 18)
    CAPTURING = ClassProperty('Capturing', 1 << 19, 19)
    OBJECT_TRANSFER = ClassProperty('Object Transfer', 1 << 20, 20)
    AUDIO = ClassProperty('Audio', 1 << 21, 21)
    TELEPHONY = ClassProperty('Telephony', 1 << 22, 22)
    INFORMATION = ClassProperty('Information', 1 << 23, 23)



import os
from contextlib import contextmanager
from datetime import datetime as dt
from typing import Optional, Tuple

from tornado.web import stream_request_body

from platypush.utils import get_mime_type

from .. import StreamingRoute


@stream_request_body
class FileRoute(StreamingRoute):
    """
    Generic route to read the content of a file on the server.
    """

    BUFSIZE = 1024

    @classmethod
    def path(cls) -> str:
        """
        Route: GET /file?path=<path>[&download]
        """
        return r"^/file$"

    @property
    def download(self) -> bool:
        return 'download' in self.request.arguments

    @property
    def file_path(self) -> str:
        return os.path.expanduser(
            self.request.arguments.get('path', [b''])[0].decode('utf-8')
        )

    @property
    def file_size(self) -> int:
        return os.path.getsize(self.file_path)

    @property
    def range(self) -> Tuple[Optional[int], Optional[int]]:
        range_hdr = self.request.headers.get('Range')
        if not range_hdr:
            return None, None

        start, end = range_hdr.split('=')[-1].split('-')
        start = int(start) if start else 0
        end = int(end) if end else self.file_size - 1
        return start, end

    def set_headers(self):
        self.set_header('Content-Length', str(os.path.getsize(self.file_path)))
        self.set_header(
            'Content-Type', get_mime_type(self.file_path) or 'application/octet-stream'
        )
        self.set_header('Accept-Ranges', 'bytes')
        self.set_header(
            'Last-Modified',
            dt.fromtimestamp(os.path.getmtime(self.file_path)).strftime(
                '%a, %d %b %Y %H:%M:%S GMT'
            ),
        )

        if self.download:
            self.set_header(
                'Content-Disposition',
                f'attachment; filename="{os.path.basename(self.file_path)}"',
            )

        if self.range[0] is not None:
            start, end = self.range
            self.set_header(
                'Content-Range',
                f'bytes {start}-{end}/{self.file_size}',
            )
            self.set_status(206)

    @contextmanager
    def _serve(self):
        path = self.file_path
        if not path:
            self.write_error(400, 'Missing path argument')
            return

        self.logger.debug('Received file read request for %r', path)

        try:
            with open(path, 'rb') as f:
                self.set_headers()
                yield f
        except FileNotFoundError:
            self.write_error(404, 'File not found')
            yield
            return
        except PermissionError:
            self.write_error(403, 'Permission denied')
            yield
            return
        except Exception as e:
            self.write_error(500, str(e))
            yield
            return

        self.finish()

    def get(self) -> None:
        with self._serve() as f:
            if f:
                while True:
                    chunk = f.read(self.BUFSIZE)
                    if not chunk:
                        break

                    self.write(chunk)
                    self.flush()

    def head(self) -> None:
        with self._serve():
            pass



from abc import ABC, abstractmethod
import asyncio
from asyncio.subprocess import PIPE
from logging import getLogger
from queue import Empty, Queue
from threading import Event, RLock, Thread
from typing import Any, Callable, Coroutine, Iterable, Optional

from platypush.context import get_or_create_event_loop
from platypush.utils import is_debug_enabled

dtype_to_ffmpeg_format = {
    'int8': 's8',
    'uint8': 'u8',
    'int16': 's16le',
    'uint16': 'u16le',
    'int32': 's32le',
    'uint32': 'u32le',
    'float32': 'f32le',
    'float64': 'f64le',
}
"""
Supported raw types:
    'int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32', 'float32', 'float64'
"""


class AudioConverter(Thread, ABC):
    """
    Base class for an ffmpeg audio converter instance.
    """

    _format_to_ffmpeg_args = {
        'wav': ('-f', 'wav'),
        'ogg': ('-f', 'ogg'),
        'mp3': ('-f', 'mp3'),
        'aac': ('-f', 'adts'),
        'flac': ('-f', 'flac'),
    }

    def __init__(
        self,
        *args,
        ffmpeg_bin: str,
        sample_rate: int,
        channels: int,
        volume: float,
        dtype: str,
        chunk_size: int,
        input_format: Optional[str] = None,  # pylint: disable=redefined-builtin
        output_format: Optional[str] = None,  # pylint: disable=redefined-builtin
        on_exit: Optional[Callable[[], Any]] = None,
        **kwargs,
    ):
        """
        :param ffmpeg_bin: Path to the ffmpeg binary.
        :param sample_rate: The sample rate of the input/output audio.
        :param channels: The number of channels of the input/output audio.
        :param volume: Audio volume, as a percentage between 0 and 100.
        :param dtype: The (numpy) data type of the raw input/output audio.
        :param chunk_size: Number of bytes that will be read at once from the
            ffmpeg process.
        :param input_format: Input audio format.
        :param output_format: Output audio format.
        :param on_exit: Function to call when the ffmpeg process exits.
        """
        super().__init__(*args, **kwargs)

        self._ffmpeg_bin = ffmpeg_bin
        self._ffmpeg_task: Optional[Coroutine] = None
        self._sample_rate = sample_rate
        self._channels = channels
        self._chunk_size = chunk_size
        self._input_format = input_format
        self._output_format = output_format
        self._dtype = dtype
        self._closed = False
        self._out_queue = Queue()
        self.ffmpeg = None
        self.volume = volume
        self.logger = getLogger(__name__)
        self._loop = None
        self._should_stop = Event()
        self._stop_lock = RLock()
        self._on_exit = on_exit
        self._ffmpeg_terminated = Event()

    def __enter__(self) -> "AudioConverter":
        """
        Audio converter context manager.

        It starts and registers the ffmpeg converter process.
        """
        self.start()
        return self

    def __exit__(self, *_, **__):
        """
        Audio converter context manager.

        It stops and unregisters the ffmpeg converter process.
        """
        self.stop()

    def _check_ffmpeg(self):
        assert not self.terminated, 'The ffmpeg process has already terminated'

    @property
    def gain(self) -> float:
        return self.volume / 100

    @property
    def terminated(self) -> bool:
        return self._ffmpeg_terminated.is_set()

    @property
    def _default_args(self) -> Iterable[str]:
        """
        Set of arguments common to all ffmpeg converter instances.
        """
        log_level = 'debug' if is_debug_enabled() else 'warning'
        return ('-hide_banner', '-loglevel', log_level, '-y')

    @property
    @abstractmethod
    def _input_format_args(self) -> Iterable[str]:
        """
        Ffmpeg audio input arguments.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def _output_format_args(self):
        """
        Ffmpeg audio output arguments.
        """
        raise NotImplementedError()

    @property
    def _channel_layout_args(self) -> Iterable[str]:
        """
        Set of extra ffmpeg arguments for the channel layout.
        """
        args = ('-ac', str(self._channels))
        if self._channels == 1:
            return args + ('-channel_layout', 'mono')
        if self._channels == 2:
            return args + ('-channel_layout', 'stereo')
        return args

    @property
    def _audio_volume_args(self) -> Iterable[str]:
        """
        Ffmpeg audio volume arguments.
        """
        return ('-filter:a', f'volume={self.gain}')

    @property
    def _input_source_args(self) -> Iterable[str]:
        """
        Default arguments for the ffmpeg input source (default: ``-i pipe:``,
        ffmpeg will read from a pipe filled by the application).
        """
        return ('-i', 'pipe:')

    @property
    def _output_target_args(self) -> Iterable[str]:
        """
        Default arguments for the ffmpeg output target (default: ``pipe:``,
        ffmpeg will write the output to a pipe read by the application).
        """
        return ('pipe:',)

    @property
    def _converter_stdin(self) -> Optional[int]:
        """
        Default stdin file descriptor to be used by the ffmpeg converter.

        Default: ``PIPE``, as the ffmpeg process by default reads audio frames
        from the stdin.
        """
        return PIPE

    async def _audio_proxy(self, timeout: Optional[float] = None):
        """
        Proxy the converted audio stream to the output queue for downstream
        consumption.
        """
        ffmpeg_args = (
            self._ffmpeg_bin,
            *self._default_args,
            *self._input_format_args,
            *self._input_source_args,
            *self._output_format_args,
            *self._output_target_args,
        )

        self.ffmpeg = await asyncio.create_subprocess_exec(
            *ffmpeg_args,
            stdin=self._converter_stdin,
            stdout=PIPE,
        )

        self.logger.info('Running ffmpeg: %s', ' '.join(ffmpeg_args))

        while (
            self._loop
            and self.ffmpeg
            and self.ffmpeg.returncode is None
            and not self.should_stop
        ):
            self._check_ffmpeg()
            assert (
                self.ffmpeg and self.ffmpeg.stdout
            ), 'The stdout is closed for the ffmpeg process'

            self._ffmpeg_terminated.clear()

            try:
                reader = asyncio.create_task(self.ffmpeg.stdout.read(self._chunk_size))
                data = await asyncio.wait_for(reader, timeout)
                self._out_queue.put(data)
            except asyncio.TimeoutError:
                pass
            except Exception as e:
                self.logger.warning('Audio proxy error: %s', e)
                break

        self._out_queue.put(b'')

    def write(self, data: bytes):
        """
        Write raw data to the ffmpeg process.
        """
        self._check_ffmpeg()
        assert (
            self.ffmpeg and self._loop and self.ffmpeg.stdin
        ), 'The stdin is closed for the ffmpeg process'

        self._loop.call_soon_threadsafe(self.ffmpeg.stdin.write, data)

    def read(self, timeout: Optional[float] = None) -> Optional[bytes]:
        """
        Read the next chunk of converted audio bytes from the converter queue.
        """
        try:
            return self._out_queue.get(timeout=timeout)
        except Empty:
            return None

    def run(self):
        """
        Main runner. It runs the audio proxy in a loop and cleans up everything
        in case of stop/failure.
        """
        super().run()
        self._loop = get_or_create_event_loop()
        try:
            self._ffmpeg_task = self._audio_proxy(timeout=1)
            self._loop.run_until_complete(self._ffmpeg_task)
        except RuntimeError as e:
            self.logger.warning(e)
        except Exception as e:
            self.logger.warning('Audio converter error: %s', e)
            self.logger.exception(e)
        finally:
            self.stop()

    def stop(self):
        """
        Sets the stop event, kills the ffmpeg process and resets the context.
        """
        with self._stop_lock:
            self._should_stop.set()
            if self._ffmpeg_task:
                self._ffmpeg_task.close()
                self._ffmpeg_task = None

            try:
                if self.ffmpeg and self.ffmpeg.returncode is None:
                    self.ffmpeg.kill()
            except ProcessLookupError:
                pass

            self.ffmpeg = None
            self._loop = None

        self._ffmpeg_terminated.set()

        if self._on_exit:
            self._on_exit()

    @property
    def should_stop(self) -> bool:
        """
        Proxy property for the ``_should_stop`` event.
        """
        return self._should_stop.is_set()


# vim:sw=4:ts=4:et:



from flask import Blueprint, jsonify, send_from_directory

from platypush.config import Config
from platypush.backend.http.app import template_folder

pwa = Blueprint('pwa', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    pwa,
]


@pwa.route('/manifest.json', methods=['GET'])
def manifest_json():
    """Generated manifest file for the PWA"""
    return jsonify(
        {
            "name": f'Platypush @ {Config.get("device_id")}',
            "short_name": Config.get('device_id'),
            "icons": [
                {
                    "src": "/img/icons/favicon-16x16.png",
                    "sizes": "16x16",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/favicon-32x32.png",
                    "sizes": "32x32",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/apple-touch-icon-60x60.png",
                    "sizes": "60x60",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/apple-touch-icon-76x76.png",
                    "sizes": "76x76",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/apple-touch-icon-120x120.png",
                    "sizes": "120x120",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/msapplication-icon-144x144.png",
                    "sizes": "144x144",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/mstile-150x150.png",
                    "sizes": "150x150",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/apple-touch-icon-152x152.png",
                    "sizes": "152x152",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/apple-touch-icon-180x180.png",
                    "sizes": "180x180",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/android-chrome-192x192.png",
                    "sizes": "192x192",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/android-chrome-maskable-192x192.png",
                    "sizes": "192x192",
                    "type": "image/png",
                    "purpose": "maskable",
                },
                {
                    "src": "/img/icons/logo-256x256.png",
                    "sizes": "256x256",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/android-chrome-512x512.png",
                    "sizes": "512x512",
                    "type": "image/png",
                },
                {
                    "src": "/img/icons/android-chrome-maskable-512x512.png",
                    "sizes": "512x512",
                    "type": "image/png",
                    "purpose": "maskable",
                },
            ],
            "gcm_sender_id": "",
            "gcm_user_visible_only": True,
            "start_url": "/",
            "permissions": ["gcm"],
            "orientation": "portrait",
            "display": "standalone",
            "theme_color": "#000000",
            "background_color": "#ffffff",
        }
    )


@pwa.route('/service-worker.js', methods=['GET'])
def service_worker_js():
    """URL that serves the service worker for the PWA"""
    return send_from_directory(template_folder, 'service-worker.js')


# vim:sw=4:ts=4:et:



from marshmallow import fields
from marshmallow.schema import Schema

from platypush.plugins.media import DownloadState
from platypush.schemas import DateTime


class MediaDownloadSchema(Schema):
    """
    Media download schema.
    """

    url = fields.URL(
        required=True,
        metadata={
            "description": "Download URL",
            "example": "https://example.com/video.mp4",
        },
    )

    path = fields.String(
        required=True,
        metadata={
            "description": "Download path",
            "example": "/path/to/download/video.mp4",
        },
    )

    state = fields.Enum(
        DownloadState,
        required=True,
        metadata={
            "description": "Download state",
        },
    )

    size = fields.Integer(
        nullable=True,
        metadata={
            "description": "Download size (bytes)",
            "example": 1024,
        },
    )

    timeout = fields.Integer(
        nullable=True,
        metadata={
            "description": "Download timeout (seconds)",
            "example": 60,
        },
    )

    started_at = DateTime(
        nullable=True,
        metadata={
            "description": "Download start time",
        },
    )

    ended_at = DateTime(
        nullable=True,
        metadata={
            "description": "Download end time",
        },
    )



from typing import Dict, Iterable, List, Optional, Union

from platypush.plugins import RunnablePlugin, action
from platypush.plugins.media import PlayerState
from platypush.schemas.mopidy import (
    MopidyAlbumSchema,
    MopidyArtistSchema,
    MopidyDirectorySchema,
    MopidyFilterSchema,
    MopidyPlaylistSchema,
    MopidyStatusSchema,
    MopidyTrackSchema,
)
from platypush.utils import wait_for_either

from ._client import MopidyClient
from ._common import DEFAULT_TIMEOUT
from ._conf import MopidyConfig
from ._exc import EmptyTrackException
from ._playlist import MopidyPlaylist
from ._status import MopidyStatus
from ._sync import PlaylistSync
from ._task import MopidyTask
from ._track import MopidyTrack


class MusicMopidyPlugin(RunnablePlugin):
    """
    This plugin allows you to track the events from a Mopidy instance
    and control it through the Mopidy HTTP API.

    Requires:

        * A Mopidy instance running with the HTTP service enabled.

    """

    def __init__(
        self,
        host: str = 'localhost',
        port: int = 6680,
        ssl: bool = False,
        timeout: Optional[float] = DEFAULT_TIMEOUT,
        **kwargs,
    ):
        """
        :param host: Mopidy host (default: localhost).
        :param port: Mopidy HTTP port (default: 6680).
        :param ssl: Set to True if the Mopidy server is running on HTTPS.
        :param timeout: Default timeout for the Mopidy requests (default: 20s).
        """
        super().__init__(**kwargs)

        self.config = MopidyConfig(host=host, port=port, ssl=ssl, timeout=timeout)
        self._status = MopidyStatus()
        self._tasks: Dict[int, MopidyTask] = {}
        self._client: Optional[MopidyClient] = None
        self._playlist_sync = PlaylistSync()

    def _exec(self, *msgs: dict, **kwargs):
        assert self._client, "Mopidy client not running"
        return self._client.exec(
            *msgs, timeout=kwargs.pop('timeout', self.config.timeout)
        )

    def _exec_with_status(self, *msgs: dict, **kwargs):
        self._exec(*msgs, **kwargs)
        return self._dump_status()

    def _dump_status(self):
        assert self._client, "Mopidy client not running"
        return MopidyStatusSchema().dump(self._client.status)

    def _dump_results(self, results: List[dict]) -> List[dict]:
        schema_by_type = {
            'artist': MopidyArtistSchema(),
            'album': MopidyAlbumSchema(),
            'directory': MopidyDirectorySchema(),
            'playlist': MopidyPlaylistSchema(),
            'track': MopidyTrackSchema(),
        }

        return [
            {
                **(
                    MopidyTrack.parse(item).to_dict()  # type: ignore
                    if item['type'] == 'track'
                    else schema_by_type[item['type']].dump(item)
                ),
                'type': item['type'],
            }
            for item in results
        ]

    def _dump_search_results(self, results: List[dict]) -> List[dict]:
        return self._dump_results(
            [
                {
                    **item,
                    'type': res_type,
                }
                for search_provider in results
                for res_type in ['artist', 'album', 'track']
                for item in search_provider.get(res_type + 's', [])
            ]
        )

    def _lookup(self, *uris: str) -> Dict[str, List[MopidyTrack]]:
        if not uris:
            return {}

        if len(uris) > 1:
            # If more than one URI is specified, we need to call only
            # library.lookup, as playlist.lookup does not support multiple
            # URIs.
            result = self._exec(
                {'method': 'core.library.lookup', 'uris': uris},
            )[0]
        else:
            # Otherwise, search both in the library and in the playlist
            # controllers.
            uri = uris[0]
            result = self._exec(
                {'method': 'core.playlists.lookup', 'uri': uri},
                {'method': 'core.library.lookup', 'uris': [uri]},
            )
            result = {
                uri: (
                    result[0].get('tracks', [])
                    if result[0]
                    else list(result[1].values())[0]
                )
            }

        ret = {}
        for uri, tracks in result.items():
            ret[uri] = []
            for track in tracks:
                parsed_track = MopidyTrack.parse(track)
                if parsed_track:
                    ret[uri].append(parsed_track)

        return ret

    def _add(
        self,
        *resources: str,
        position: Optional[int] = None,
        clear: bool = False,
        lookup: bool = True,
    ):
        batch_size = 50
        results = self._lookup(*resources).values()
        ret = []
        uris = (
            [track.uri for tracks in results for track in tracks if track and track.uri]
            if lookup
            else list(resources)
        )

        with self._playlist_sync:
            if clear:
                self.clear()

            for i in range(0, len(uris), batch_size):
                ret += self._exec(
                    {
                        'method': 'core.tracklist.add',
                        'uris': uris[i : i + batch_size],
                        'at_position': position,
                    }
                )[0]

                self.logger.info('Loaded %d/%d tracks', len(ret), len(uris))

        return ret

    def _get_playlist(self, playlist: str, with_tracks: bool = False) -> MopidyPlaylist:
        playlists = self._get_playlists()
        pl_by_name = {p.name: p for p in playlists}
        pl_by_uri = {p.uri: p for p in playlists}
        pl = pl_by_uri.get(playlist, pl_by_name.get(playlist))
        assert pl, f"Playlist {playlist} not found"

        if with_tracks:
            pl.tracks = self._get_playlist_tracks(playlist)

        return pl

    def _get_playlist_tracks(self, playlist: str) -> List[MopidyTrack]:
        playlists = self._get_playlists()
        pl_by_name = {p.name: p for p in playlists}
        pl_by_uri = {p.uri: p for p in playlists}
        pl = pl_by_uri.get(playlist, pl_by_name.get(playlist))
        assert pl, f"Playlist {playlist} not found"

        tracks = self._exec({'method': 'core.playlists.get_items', 'uri': pl.uri})[0]
        assert tracks is not None, f"Playlist {playlist} not found"

        ret = []
        for track in tracks:
            parsed_track = MopidyTrack.parse(track)
            if parsed_track:
                ret.append(parsed_track)

        return ret

    def _get_playlists(self, **__) -> List[MopidyPlaylist]:
        return [
            MopidyPlaylist.parse(pl)
            for pl in self._exec({'method': 'core.playlists.as_list'})[0]
        ]

    def _save_playlist(self, playlist: MopidyPlaylist):
        return self._exec(
            {
                'method': 'core.playlists.save',
                'playlist': {
                    '__model__': 'Playlist',
                    'uri': playlist.uri,
                    'name': playlist.name,
                    'tracks': [
                        {
                            '__model__': 'Track',
                            'uri': track.uri,
                        }
                        for track in playlist.tracks
                    ],
                },
            }
        )[0]

    @action
    def play(
        self,
        resource: Optional[str] = None,
        position: Optional[int] = None,
        track_id: Optional[int] = None,
        **__,
    ):
        """
        Start playback, or play a resource by URI.

        :param resource: Resource path/URI. If not specified, it will resume the
            playback if paused/stopped, otherwise it will start playing the
            selected track.
        :param track_id: The ID of track (or ``tlid``) in the current playlist
            that should be played, if you want to play a specific track already
            loaded in the current playlist.
        :param position: Position number (0-based) of the track in the current
            playlist that should be played.
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if resource:
            ret = self._add(resource, position=0)
            if not ret:
                self.logger.warning('Failed to add %s to the tracklist', resource)
            elif isinstance(ret, list):
                track_id = ret[0].get('tlid')
            elif isinstance(ret, dict):
                track_id = ret.get('tlid')
        elif position is not None:
            tracklist = self._exec({'method': 'core.tracklist.get_tl_tracks'})[0]
            if position < 0 or position >= len(tracklist):
                self.logger.warning(
                    'Position %d is out of bounds for the current tracklist', position
                )
                return None

            track_id = tracklist[position]['tlid']

        return self._exec_with_status(
            {'method': 'core.playback.play', 'tlid': track_id}
        )

    @action
    def play_pos(self, pos: int):
        """
        Play a track in the current playlist by position number.

        Legacy alias for :meth:`.play` with a ``position`` parameter.

        :param pos: Position number (0-based).
        """
        return self.play(position=pos)

    @action
    def load(self, playlist: str, play: bool = True):
        """
        Load/play a playlist.

        This method will clear the current playlist and load the tracks from the
        given playlist.

        You should usually prefer :meth:`.add` to this method, as it is more
        general-purpose (``load`` only works with playlists). This method exists
        mainly for compatibility with the MPD plugin.

        :param playlist: Playlist URI.
        :param play: Start playback after loading the playlist (default: True).
        """
        self._add(playlist, clear=True)
        if play:
            self.play()

    @action
    def lookup(self, resources: Iterable[str], **__):
        """
        Lookup (one or) resources by URI.

        Given a list of URIs, this method will return a dictionary in the form
        ``{uri: [track1, track2, ...]}``.

        :param resource: Resource URI(s).
        :return: .. schema:: mopidy.MopidyTrackSchema(many=True)
        """
        return {
            uri: [track.to_dict() for track in tracks]
            for uri, tracks in self._lookup(*resources).items()
        }

    @action
    def pause(self, **__):
        """
        Pause the playback.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        return self._exec_with_status({'method': 'core.playback.pause'})

    @action
    def stop(self, **__):  # type: ignore
        """
        Stop the playback.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        # Note: stop could be interpreted both as "stop the playback" and "stop
        # the plugin". If the stop event is set, we assume that the user wants
        # to stop the plugin.
        if self.should_stop():
            if self._client:
                self._client.stop()

            return None

        return self._exec_with_status({'method': 'core.playback.stop'})

    @action
    def prev(self, **__):
        """
        Play the previous track.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        return self._exec_with_status({'method': 'core.playback.previous'})

    @action
    def next(self, **__):
        """
        Play the next track.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        return self._exec_with_status({'method': 'core.playback.next'})

    @action
    def add(
        self,
        resource: Union[str, Iterable[str]],
        *_,
        position: Optional[int] = None,
        **__,
    ):
        """
        Add a resource (track, album, artist, folder etc.) to the current
        playlist.

        :param resource: Resource URI(s).
        :param position: Position (0-based) where the track(s) will be inserted
            (default: end of the playlist).
        :return: The list of tracks added to the queue.
            .. schema:: mopidy.MopidyTrackSchema(many=True)
        """
        resources = [resource] if isinstance(resource, str) else resource
        tracks = [
            MopidyTrack.parse(track)
            for track in self._add(*resources, position=position)
        ]
        return [track.to_dict() for track in tracks if track]

    @action
    def pause_if_playing(self, **__):
        """
        Pause the playback if it's currently playing.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if self._status.state == PlayerState.PLAY:
            return self.pause()

        return self._dump_status()

    @action
    def play_if_paused(self, **__):
        """
        Resume the playback if it's currently paused.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if self._status.state == PlayerState.PAUSE:
            return self.play()

        return self._dump_status()

    @action
    def play_if_paused_or_stopped(self):
        """
        Resume the playback if it's currently paused or stopped.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if self._status.state in {PlayerState.PAUSE, PlayerState.STOP}:
            return self.play()

        return self._dump_status()

    @action
    def play_or_stop(self):
        """
        Play if the playback is stopped, stop if it's playing, otherwise resume
        playback.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if self._status.state == PlayerState.PLAY:
            return self.stop()

        return self.play()

    @action
    def set_volume(self, volume: int, **__):
        """
        Set the volume.

        :param volume: Volume level (0-100).
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        return self._exec_with_status(
            {'method': 'core.mixer.set_volume', 'volume': volume}
        )

    @action
    def volup(self, step: int = 5, **__):
        """
        Increase the volume by a given step.

        :param step: Volume step (default: 5%).
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        return self.set_volume(volume=min(100, self._status.volume + step))

    @action
    def voldown(self, step: int = 5, **__):
        """
        Decrease the volume by a given step.

        :param step: Volume step (default: 5%).
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        return self.set_volume(volume=max(0, self._status.volume - step))

    @action
    def random(self, value: Optional[bool] = None, **__):
        """
        Set the random mode.

        :param value: Random mode. If not specified, it will toggle the current
            random mode.
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if value is None:
            value = not self._status.random

        return self._exec_with_status(
            {'method': 'core.tracklist.set_random', 'value': bool(value)}
        )

    @action
    def repeat(self, value: Optional[bool] = None, **__):
        """
        Set the repeat mode.

        :param value: Repeat mode. If not specified, it will toggle the current
            repeat mode.
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if value is None:
            value = not self._status.repeat

        return self._exec_with_status(
            {'method': 'core.tracklist.set_repeat', 'value': bool(value)}
        )

    @action
    def consume(self, value: Optional[bool] = None, **__):
        """
        Set the consume mode.

        :param value: Consume mode. If not specified, it will toggle the current
            consume mode.
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if value is None:
            value = not self._status.consume

        return self._exec_with_status(
            {'method': 'core.tracklist.set_consume', 'value': bool(value)}
        )

    @action
    def single(self, value: Optional[bool] = None, **__):
        """
        Set the single mode.

        :param value: Single mode. If not specified, it will toggle the current
            single mode.
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if value is None:
            value = not self._status.single

        return self._exec_with_status(
            {'method': 'core.tracklist.set_single', 'value': bool(value)}
        )

    @action
    def shuffle(self, **__):
        """
        Shuffle the current playlist.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        return self._exec_with_status({'method': 'core.tracklist.shuffle'})

    @action
    def save(self, name: str, **__):
        """
        Save the current tracklist to a new playlist with the given name.

        :param name: New playlist name.
        """
        return self._exec({'method': 'core.playlists.save', 'name': name})[0]

    @action
    def delete(
        self,
        positions: Optional[Iterable[int]] = None,
        uris: Optional[Iterable[str]] = None,
    ):
        """
        Delete tracks from the current tracklist.

        .. note:: At least one of the ``positions`` or ``uris`` parameters must
                  be specified.

        :param positions: (0-based) positions of the tracks to be deleted.
        :param uris: URIs of the tracks to be deleted.
        """
        assert (
            positions or uris
        ), "At least one of 'positions' or 'uris' must be specified"
        criteria = {}
        if positions:
            assert self._client, "Mopidy client not running"
            positions = set(positions)
            criteria['tlid'] = list(
                {
                    track.track_id
                    for i, track in enumerate(self._client.tracks)
                    if i in positions
                }
            )
        if uris:
            criteria['uri'] = list(uris)

        return self._exec(
            {
                'method': 'core.tracklist.remove',
                'criteria': criteria,
            }
        )[0]

    @action
    def move(
        self,
        start: Optional[int] = None,
        end: Optional[int] = None,
        position: Optional[int] = None,
        from_pos: Optional[int] = None,
        to_pos: Optional[int] = None,
        **__,
    ):
        """
        Move one or more tracks in the current playlist to a new position.

        You can pass either:

            - ``start``, ``end`` and ``position`` to move a slice of tracks
              from ``start`` to ``end`` to the new position ``position``.
            - ``from_pos`` and ``to_pos`` to move a single track from
              ``from_pos`` to ``to_pos``.

        .. note: Positions are 0-based (i.e. the first track has position 0).

        :param start: Start position of the slice of tracks to be moved.
        :param end: End position of the slice of tracks to be moved.
        :param position: New position where the tracks will be inserted.
        :param from_pos: Alias for ``start`` - it only works with one track at
            the time. Maintained for compatibility with
            :meth:`platypush.plugins.music.mpd.MusicMpdPlugin.move`.
        :param to_pos: Alias for ``position`` - it only works with one track at
            the time. Maintained for compatibility with
            :meth:`platypush.plugins.music.mpd.MusicMpdPlugin.move`.
        """
        assert (from_pos is not None and to_pos is not None) or (
            start is not None and end is not None and position is not None
        ), 'Either "start", "end" and "position", or "from_pos" and "to_pos" must be specified'

        if (from_pos is not None) and (to_pos is not None):
            start, end, position = from_pos, from_pos, to_pos

        ret = self._exec(
            {
                'method': 'core.tracklist.move',
                'start': start,
                'end': end,
                'to_position': position,
            }
        )[0]

        if self._client:
            self._client.refresh_status(with_tracks=True)
        return ret

    @action
    def clear(self, **__):
        """
        Clear the current playlist.
        """
        self._exec_with_status({'method': 'core.tracklist.clear'})

    @action
    def seek(self, position: float, **__):
        """
        Seek to a given position in the current track.

        :param position: Position in seconds.
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        return self._exec_with_status(
            {'method': 'core.playback.seek', 'time_position': int(position * 1000)}
        )

    @action
    def back(self, delta: float = 10, **__):
        """
        Seek back by a given number of seconds.

        :param delta: Number of seconds to seek back (default: 10s).
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if self._status.playing_pos is None:
            return self._dump_status()

        return self.seek(position=self._status.playing_pos - delta)

    @action
    def forward(self, delta: float = 10, **__):
        """
        Seek forward by a given number of seconds.

        :param delta: Number of seconds to seek forward (default: 10s).
        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        if self._status.playing_pos is None:
            return self._dump_status()

        return self.seek(position=self._status.playing_pos + delta)

    @action
    def status(self, **__):
        """
        Get the current Mopidy status.

        :return: .. schema:: mopidy.MopidyStatusSchema
        """
        assert self._client, "Mopidy client not running"
        self._client.refresh_status()
        return self._dump_status()

    @action
    def current_track(self, **__):
        """
        Get the current track.

        :return: .. schema:: mopidy.MopidyTrackSchema
        """
        assert self._client, "Mopidy client not running"
        if not self._client.status.track:
            return None

        return self._client.status.track.to_dict()

    @action
    def get_tracks(self, **__):
        """
        Get the current playlist tracks.

        :return: .. schema:: mopidy.MopidyTrackSchema(many=True)
        """
        assert self._client, "Mopidy client not running"
        return [t.to_dict() for t in self._client.tracks]

    @action
    def get_playlists(self, **__):
        """
        Get the available playlists.

        :return: .. schema:: mopidy.MopidyPlaylistSchema(many=True)
        """
        return MopidyPlaylistSchema().dump(self._get_playlists(), many=True)

    @action
    def get_playlist(self, playlist: str, **__):
        """
        Get the items in a playlist.

        :param playlist: Playlist URI.
        :param only_tracks: If True, only the tracks will be returned, otherwise
            the full playlist object will be returned - including name and other
            metadata.
        :return: .. schema:: mopidy.MopidyTrackSchema(many=True)
        """
        tracks = self._get_playlist_tracks(playlist)
        tracks_by_uri = {t.uri: t for t in tracks if t.uri}
        looked_up = self._lookup(*tracks_by_uri.keys())
        return [
            track.to_dict()
            for track in [
                (looked_up[uri][0] if looked_up.get(uri) else track)
                for uri, track in tracks_by_uri.items()
            ]
        ]

    @action
    def get_playlist_uri_schemes(self, **__):
        """
        Get the available playlist URI schemes.

        :return: List of available playlist URI schemes.
        """
        return self._exec({'method': 'core.playlists.get_uri_schemes'})[0]

    @action
    def create_playlist(self, name: str, uri_scheme: str = 'm3u', **__):
        """
        Create a new playlist.

        :param name: Playlist name.
        :param uri_scheme: URI scheme for the playlist (default: ``m3u``).
            You can get a full list of the available URI schemes that support
            playlist creation on the Mopidy instance by calling
            :meth:`.get_playlist_uri_schemes`.
        :return: .. schema:: mopidy.MopidyPlaylistSchema
        """
        return MopidyPlaylistSchema().dump(
            self._exec(
                {
                    'method': 'core.playlists.create',
                    'name': name,
                    'uri_scheme': uri_scheme,
                }
            )[0]
        )

    @action
    def delete_playlist(self, playlist: str, **__):
        """
        Delete a playlist.

        :param playlist: Playlist URI.
        :return: ``True`` if the playlist was deleted, ``False`` otherwise.
        """
        return self._exec({'method': 'core.playlists.delete', 'uri': playlist})[0]

    @action
    def add_to_playlist(
        self,
        playlist: str,
        resources: Iterable[str],
        position: Optional[int] = None,
        allow_duplicates: bool = False,
        **__,
    ):
        """
        Add tracks to a playlist.

        :param playlist: Playlist URI/name.
        :param resources: List of track URIs.
        :param position: Position where the tracks will be inserted (default:
            end of the playlist).
        :param allow_duplicates: If True, the tracks will be added even if they
            are already present in the playlist (default: False).
        :return: The modified playlist.
            .. schema:: mopidy.MopidyPlaylistSchema
        """
        pl = self._get_playlist(playlist, with_tracks=True)

        if not allow_duplicates:
            existing_uris = {t.uri for t in pl.tracks}
            resources = [t for t in resources if t not in existing_uris]

        new_tracks = [MopidyTrack(uri=t) for t in resources]
        if position is not None:
            pl.tracks = pl.tracks[:position] + new_tracks + pl.tracks[position:]
        else:
            pl.tracks += new_tracks

        self._save_playlist(pl)
        return pl.to_dict()

    @action
    def remove_from_playlist(
        self,
        playlist: str,
        resources: Optional[Iterable[Union[str, int]]] = None,
        from_pos: Optional[int] = None,
        to_pos: Optional[int] = None,
        **__,
    ):
        """
        Remove tracks from a playlist.

        This action can work in three different ways:

            - If the ``resources`` parameter is specified, and it contains
              strings, it will remove the tracks matching the provided URIs.
            - If the ``resources`` parameter is specified, and it contains
              integers, it will remove the tracks in the specified positions.
            - If the ``from_pos`` and ``to_pos`` parameters are specified, it
              will remove the tracks in the specified range (inclusive).

        .. note: Positions are 0-based (i.e. the first track has position 0).

        :param playlist: Playlist URI/name.
        :param tracks: List of track URIs.
        :param from_pos: Start position of the slice of tracks to be removed.
        :param to_pos: End position of the slice of tracks to be removed.
        :return: The modified playlist.
            .. schema:: mopidy.MopidyPlaylistSchema
        """
        assert resources or (
            from_pos is not None and to_pos is not None
        ), "Either 'tracks', or 'positions', or 'from_pos' and 'to_pos' must be specified"

        pl = self._get_playlist(playlist, with_tracks=True)

        if resources:
            resources = set(resources)
            positions = {
                i
                for i, t in enumerate(pl.tracks)
                if t.uri in resources or i in resources
            }

            pl.tracks = [t for i, t in enumerate(pl.tracks) if i not in positions]
        elif from_pos is not None and to_pos is not None:
            from_pos, to_pos = (min(from_pos, to_pos), max(from_pos, to_pos))
            pl.tracks = pl.tracks[: from_pos - 1] + pl.tracks[to_pos:]

        self._save_playlist(pl)
        return pl.to_dict()

    @action
    def playlist_move(
        self,
        playlist: str,
        start: Optional[int] = None,
        end: Optional[int] = None,
        position: Optional[int] = None,
        from_pos: Optional[int] = None,
        to_pos: Optional[int] = None,
        **__,
    ):
        """
        Move tracks in a playlist.

        This action can work in two different ways:

            - If the ``start``, ``end`` and ``position`` parameters are
              specified, it will move an individual track from the position
              ``start`` to the position ``end`` to the new position
              ``position``.

            - If the ``from_pos``, ``to_pos`` and ``position`` parameters are
              specified, it will move the tracks in the specified range
              (inclusive) to the new position ``position``.

        .. note: Positions are 0-based (i.e. the first track has position 0).

        :param playlist: Playlist URI.
        :param start: Start position of the slice of tracks to be moved.
        :param end: End position of the slice of tracks to be moved.
        :param position: New position where the tracks will be inserted.
        :return: The modified playlist.
            .. schema:: mopidy.MopidyPlaylistSchema
        """
        assert (start is not None and end is not None and position is not None) or (
            from_pos is not None and to_pos is not None
        ), "Either 'start', 'end' and 'position', or 'from_pos' and 'to_pos' must be specified"

        pl = self._get_playlist(playlist, with_tracks=True)

        if from_pos is not None and to_pos is not None:
            from_pos, to_pos = (min(from_pos, to_pos), max(from_pos, to_pos))
            pl.tracks = (
                pl.tracks[:from_pos]
                + pl.tracks[to_pos : to_pos + 1]
                + pl.tracks[from_pos + 1 : to_pos]
                + pl.tracks[from_pos : from_pos + 1]
                + pl.tracks[to_pos + 1 :]
            )
        elif start is not None and end is not None and position is not None:
            start, end = (min(start, end), max(start, end))
            if start == end:
                end += 1

            if start < position:
                pl.tracks = (
                    pl.tracks[:start]
                    + pl.tracks[end : end + (position - start)]
                    + pl.tracks[start:end]
                    + pl.tracks[end + (position - start) :]
                )
            else:
                pl.tracks = (
                    pl.tracks[:position]
                    + pl.tracks[start:end]
                    + pl.tracks[position:start]
                    + pl.tracks[end:]
                )

        self._save_playlist(pl)
        return pl.to_dict()

    @action
    def playlist_clear(self, playlist: str, **__):
        """
        Remove all the tracks from a playlist.

        :param playlist: Playlist URI/name.
        :return: The modified playlist.
            .. schema:: mopidy.MopidyPlaylistSchema
        """
        pl = self._get_playlist(playlist)
        pl.tracks = []
        self._save_playlist(pl)
        return pl.to_dict()

    @action
    def rename_playlist(self, playlist: str, new_name: str, **__):
        """
        Rename a playlist.

        :param playlist: Playlist URI/name.
        :param new_name: New playlist name.
        :return: The modified playlist.
            .. schema:: mopidy.MopidyPlaylistSchema
        """
        pl = self._get_playlist(playlist, with_tracks=True)
        pl.name = new_name
        self._save_playlist(pl)
        return pl.to_dict()

    @action
    def get_images(self, resources: Iterable[str], **__) -> Dict[str, Optional[str]]:
        """
        Get the images for a list of URIs.

        :param resources: List of URIs.
        :return: Dictionary in the form ``{uri: image_url}``.
        """
        return {
            uri: next(iter(images or []), {}).get('uri')
            for uri, images in self._exec(
                {'method': 'core.library.get_images', 'uris': list(resources)}
            )[0].items()
        }

    @action
    def search(  # pylint: disable=redefined-builtin
        self, filter: dict, exact: bool = False, **__
    ):
        """
        Search items that match the given query.

        :param filter: .. schema:: mopidy.MopidyFilterSchema
        :param exact: If True, the search will only return exact matches.
        :return: A list of result, including:

            - Tracks
                .. schema:: mopidy.MopidyTrackSchema(many=True)
            - Albums
                .. schema:: mopidy.MopidyAlbumSchema(many=True)
            - Artists
                .. schema:: mopidy.MopidyArtistSchema(many=True)

        """
        filter = dict(MopidyFilterSchema().load(filter) or {})
        uris = filter.pop('uris', None)
        kwargs = {
            'exact': exact,
            'query': filter,
            **({'uris': uris} if uris else {}),
        }

        return self._dump_search_results(
            self._exec({'method': 'core.library.search', **kwargs})[0]
        )

    @action
    def find(  # pylint: disable=redefined-builtin
        self, filter: dict, exact: bool = False, **__
    ):
        """
        Alias for :meth:`search`, for MPD compatibility.

        :param filter: .. schema:: mopidy.MopidyFilterSchema
        :param exact: If True, the search will only return exact matches.
        :return: .. schema:: mopidy.MopidyTrackSchema(many=True)
        """
        return self.search(filter=filter, exact=exact)

    @action
    def browse(self, uri: Optional[str] = None):
        """
        Browse the items under the specified URI.

        :param uri: URI to browse (default: root directory).
        :return: A list of result under the specified resource, including:

            - Directories
                .. schema:: mopidy.MopidyDirectorySchema(many=True)
            - Tracks
                .. schema:: mopidy.MopidyTrackSchema(many=True)
            - Albums
                .. schema:: mopidy.MopidyAlbumSchema(many=True)
            - Artists
                .. schema:: mopidy.MopidyArtistSchema(many=True)

        """
        return self._dump_results(
            self._exec({'method': 'core.library.browse', 'uri': uri})[0]
        )

    def main(self):
        while not self.should_stop():
            try:
                with MopidyClient(
                    config=self.config,
                    status=self._status,
                    stop_event=self._should_stop,
                    playlist_sync=self._playlist_sync,
                    tasks=self._tasks,
                ) as self._client:
                    self._client.start()
                    wait_for_either(self._should_stop, self._client.closed_event)
            finally:
                self._client = None
                self.wait_stop(10)


__all__ = ['EmptyTrackException', 'MusicMopidyPlugin', 'MopidyStatus', 'MopidyTrack']


# vim:sw=4:ts=4:et:



import logging
import os
from queue import Full, Queue
from threading import Event, RLock, Thread
from time import time
from typing import Any, Dict, Optional, Sequence

import pvporcupine

from platypush.common.assistant import AudioRecorder
from platypush.context import get_plugin
from platypush.message.event.assistant import (
    AssistantEvent,
    ConversationTimeoutEvent,
    HotwordDetectedEvent,
    IntentRecognizedEvent,
    SpeechRecognizedEvent,
)
from platypush.plugins.tts.picovoice import TtsPicovoicePlugin
from ._speech import SpeechProcessor
from ._state import AssistantState


class Assistant(Thread):
    """
    A facade class that wraps the Picovoice engines under an assistant API.
    """

    @staticmethod
    def _default_callback(*_, **__):
        pass

    def __init__(
        self,
        access_key: str,
        stop_event: Event,
        hotword_enabled: bool = True,
        stt_enabled: bool = True,
        keywords: Optional[Sequence[str]] = None,
        keyword_paths: Optional[Sequence[str]] = None,
        keyword_model_path: Optional[str] = None,
        frame_expiration: float = 3.0,  # Don't process audio frames older than this
        speech_model_path: Optional[str] = None,
        intent_model_path: Optional[str] = None,
        endpoint_duration: Optional[float] = None,
        enable_automatic_punctuation: bool = False,
        start_conversation_on_hotword: bool = False,
        audio_queue_size: int = 100,
        muted: bool = False,
        conversation_timeout: Optional[float] = None,
        on_conversation_start=_default_callback,
        on_conversation_end=_default_callback,
        on_conversation_timeout=_default_callback,
        on_speech_recognized=_default_callback,
        on_intent_matched=_default_callback,
        on_hotword_detected=_default_callback,
    ):
        super().__init__(name='picovoice:Assistant')

        self._access_key = access_key
        self._stop_event = stop_event
        self.logger = logging.getLogger(__name__)
        self.hotword_enabled = hotword_enabled
        self.stt_enabled = stt_enabled
        self.keywords = list(keywords or [])
        self.keyword_paths = None
        self.keyword_model_path = None
        self.frame_expiration = frame_expiration
        self.endpoint_duration = endpoint_duration
        self.enable_automatic_punctuation = enable_automatic_punctuation
        self.start_conversation_on_hotword = start_conversation_on_hotword
        self.audio_queue_size = audio_queue_size
        self._responding = Event()
        self._muted = muted
        self._speech_model_path = speech_model_path
        self._speech_model_path_override = None
        self._intent_model_path = intent_model_path
        self._intent_model_path_override = None
        self._in_ctx = False

        self._speech_processor = SpeechProcessor(
            stop_event=stop_event,
            stt_enabled=stt_enabled,
            intent_enabled=self.intent_enabled,
            conversation_timeout=conversation_timeout,
            model_path=speech_model_path,
            get_cheetah_args=self._get_speech_engine_args,
            get_rhino_args=self._get_intent_engine_args,
        )

        self._on_conversation_start = on_conversation_start
        self._on_conversation_end = on_conversation_end
        self._on_conversation_timeout = on_conversation_timeout
        self._on_speech_recognized = on_speech_recognized
        self._on_intent_matched = on_intent_matched
        self._on_hotword_detected = on_hotword_detected

        self._recorder = None
        self._state = AssistantState.IDLE
        self._state_lock = RLock()
        self._evt_queue = Queue(maxsize=100)

        if hotword_enabled:
            if not keywords:
                raise ValueError(
                    'You need to provide a list of keywords if the wake-word engine is enabled'
                )

            if keyword_paths:
                keyword_paths = [os.path.expanduser(path) for path in keyword_paths]
                missing_paths = [
                    path for path in keyword_paths if not os.path.isfile(path)
                ]
                if missing_paths:
                    raise FileNotFoundError(f'Keyword files not found: {missing_paths}')

                self.keyword_paths = keyword_paths

            if keyword_model_path:
                keyword_model_path = os.path.expanduser(keyword_model_path)
                if not os.path.isfile(keyword_model_path):
                    raise FileNotFoundError(
                        f'Keyword model file not found: {keyword_model_path}'
                    )

                self.keyword_model_path = keyword_model_path

        self._porcupine: Optional[pvporcupine.Porcupine] = None

    @property
    def intent_enabled(self) -> bool:
        return self.intent_model_path is not None

    @property
    def is_responding(self) -> bool:
        return self._responding.is_set()

    @property
    def speech_model_path(self) -> Optional[str]:
        return self._speech_model_path_override or self._speech_model_path

    @property
    def intent_model_path(self) -> Optional[str]:
        return self._intent_model_path_override or self._intent_model_path

    @property
    def tts(self) -> TtsPicovoicePlugin:
        p = get_plugin('tts.picovoice')
        assert p, 'Picovoice TTS plugin not configured/found'
        return p

    def set_responding(self, responding: bool):
        if responding:
            self._responding.set()
        else:
            self._responding.clear()

    def should_stop(self):
        return self._stop_event.is_set()

    def wait_stop(self):
        self._stop_event.wait()

    @property
    def state(self) -> AssistantState:
        with self._state_lock:
            return self._state

    @state.setter
    def state(self, state: AssistantState):
        with self._state_lock:
            prev_state = self._state
            self._state = state
            new_state = self.state

        if prev_state == new_state:
            return

        self.logger.info('Assistant state transition: %s -> %s', prev_state, new_state)
        if prev_state == AssistantState.DETECTING_SPEECH:
            self.tts.stop()
            self._speech_model_path_override = None
            self._intent_model_path_override = None
            self._speech_processor.on_conversation_end()
            self._on_conversation_end()
        elif new_state == AssistantState.DETECTING_SPEECH:
            self._speech_processor.on_conversation_start()
            self._on_conversation_start()

        if new_state == AssistantState.DETECTING_HOTWORD:
            self.tts.stop()
            self._speech_processor.on_conversation_reset()

        # Put a null event on the event queue to unblock next_event
        self._evt_queue.put(None)

    @property
    def porcupine(self) -> Optional[pvporcupine.Porcupine]:
        if not self.hotword_enabled:
            return None

        if not self._porcupine:
            args: Dict[str, Any] = {'access_key': self._access_key}
            if self.keywords:
                args['keywords'] = self.keywords
            if self.keyword_paths:
                args['keyword_paths'] = self.keyword_paths
            if self.keyword_model_path:
                args['model_path'] = self.keyword_model_path

            self._porcupine = pvporcupine.create(**args)

        return self._porcupine

    def _get_speech_engine_args(self) -> dict:
        args: Dict[str, Any] = {'access_key': self._access_key}
        if self.speech_model_path:
            args['model_path'] = self.speech_model_path
        if self.endpoint_duration:
            args['endpoint_duration_sec'] = self.endpoint_duration
        if self.enable_automatic_punctuation:
            args['enable_automatic_punctuation'] = self.enable_automatic_punctuation

        return args

    def _get_intent_engine_args(self) -> dict:
        args: Dict[str, Any] = {'access_key': self._access_key}
        args['context_path'] = self.intent_model_path
        if self.endpoint_duration:
            args['endpoint_duration_sec'] = self.endpoint_duration
        if self.enable_automatic_punctuation:
            args['enable_automatic_punctuation'] = self.enable_automatic_punctuation

        return args

    def __enter__(self):
        """
        Get the assistant ready to start processing audio frames.
        """
        if self.should_stop():
            return self

        assert not self.is_alive(), 'The assistant is already running'
        self._in_ctx = True

        if self._recorder:
            self.logger.info('A recording stream already exists')
        elif self.hotword_enabled or self.stt_enabled or self.intent_enabled:
            sample_rate = (self.porcupine or self._speech_processor).sample_rate
            frame_length = (self.porcupine or self._speech_processor).frame_length
            self._recorder = AudioRecorder(
                stop_event=self._stop_event,
                sample_rate=sample_rate,
                frame_size=frame_length,
                queue_size=self.audio_queue_size,
                paused=self._muted,
                channels=1,
            )

            self._speech_processor.__enter__()
            self._recorder.__enter__()

            if self.porcupine:
                self.state = AssistantState.DETECTING_HOTWORD
            else:
                self.state = AssistantState.DETECTING_SPEECH

        self.start()
        return self

    def __exit__(self, *_):
        """
        Stop the assistant and release all resources.
        """
        self._in_ctx = False
        if self._recorder:
            self._recorder.__exit__(*_)
            self._recorder = None

        self.state = AssistantState.IDLE

        if self._porcupine:
            self._porcupine.delete()
            self._porcupine = None

        self._speech_processor.__exit__(*_)

    def __iter__(self):
        """
        Iterate over processed assistant events.
        """
        return self

    def __next__(self):
        """
        Process the next audio frame and return the corresponding event.
        """
        if self.should_stop() or not self._recorder:
            raise StopIteration

        if self.hotword_enabled and self.state == AssistantState.DETECTING_HOTWORD:
            return self._evt_queue.get()

        evt = None
        if (
            self._speech_processor.enabled
            and self.state == AssistantState.DETECTING_SPEECH
        ):
            evt = self._speech_processor.next_event()

        if isinstance(evt, SpeechRecognizedEvent):
            self._on_speech_recognized(phrase=evt.args['phrase'])
        if isinstance(evt, IntentRecognizedEvent):
            self._on_intent_matched(
                intent=evt.args['intent'], slots=evt.args.get('slots', {})
            )
        if isinstance(evt, ConversationTimeoutEvent):
            self._on_conversation_timeout()

        if evt:
            self._speech_processor.reset()

        if (
            evt
            and self.state == AssistantState.DETECTING_SPEECH
            and self.hotword_enabled
        ):
            self.state = AssistantState.DETECTING_HOTWORD

        return evt

    def mute(self):
        self._muted = True
        if self._recorder:
            self._recorder.pause()

    def unmute(self):
        self._muted = False
        if self._recorder:
            self._recorder.resume()

    def set_mic_mute(self, mute: bool):
        if mute:
            self.mute()
        else:
            self.unmute()

    def toggle_mic_mute(self):
        if self._muted:
            self.unmute()
        else:
            self.mute()

    def _process_hotword(self, frame) -> Optional[HotwordDetectedEvent]:
        if not self.porcupine:
            return None

        keyword_index = self.porcupine.process(frame)
        if keyword_index is None:
            return None  # No keyword detected

        if keyword_index >= 0 and self.keywords:
            if self.start_conversation_on_hotword:
                self.state = AssistantState.DETECTING_SPEECH

            self.tts.stop()  # Stop any ongoing TTS when the hotword is detected
            self._on_hotword_detected(hotword=self.keywords[keyword_index])
            return HotwordDetectedEvent(hotword=self.keywords[keyword_index])

        return None

    def override_speech_model(self, model_path: Optional[str]):
        self._speech_model_path_override = model_path

    def override_intent_model(self, model_path: Optional[str]):
        self._intent_model_path_override = model_path

    def _put_event(self, evt: AssistantEvent):
        try:
            self._evt_queue.put_nowait(evt)
        except Full:
            self.logger.warning('The assistant event queue is full')

    def run(self):
        assert (
            self._in_ctx
        ), 'The assistant can only be started through a context manager'

        super().run()

        while not self.should_stop() and self._recorder:
            self._recorder.wait_start()
            if self.should_stop():
                break

            data = self._recorder.read()
            if data is None:
                continue

            frame, t = data
            if time() - t > self.frame_expiration:
                self.logger.info(
                    'Skipping audio frame older than %ss', self.frame_expiration
                )
                continue  # The audio frame is too old

            if self.hotword_enabled and self.state == AssistantState.DETECTING_HOTWORD:
                evt = self._process_hotword(frame)
                if evt:
                    self._put_event(evt)

                continue

            if (
                self._speech_processor.enabled
                and self.state == AssistantState.DETECTING_SPEECH
            ):
                self._speech_processor.process(frame, block=False)

        self.logger.info('Assistant stopped')


# vim:sw=4:ts=4:et:



from datetime import datetime
from dataclasses import dataclass
from typing import Optional, Dict

from platypush.message.event import Event


@dataclass
class Actor:
    id: str
    login: str
    display_login: str
    url: str
    gravatar_id: str
    avatar_url: str


@dataclass
class Repo:
    id: str
    name: str
    url: str


class GithubEvent(Event):
    """ Generic Github event """

    def __init__(self,
                 event_type: str,
                 created_at: datetime,
                 actor: Optional[Dict[str, str]] = None,
                 repo: Optional[Dict[str, str]] = None,
                 *args, **kwargs):
        super().__init__(*args, actor=actor, event_type=event_type, repo=repo, created_at=created_at, **kwargs)
        self.event_type = event_type
        self.actor = Actor(**actor) if actor else None
        self.repo = Repo(**repo) if repo else None
        self.created_at = created_at


class GithubPushEvent(GithubEvent):
    """ Github push event. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubCommitCommentEvent(GithubEvent):
    """ A commit comment is created. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubCreateEvent(GithubEvent):
    """ A git branch or tag is created. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubDeleteEvent(GithubEvent):
    """ A git branch or tag is deleted. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubForkEvent(GithubEvent):
    """ A user forks a watched repository. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubWikiEvent(GithubEvent):
    """ A wiki page is created or updated on a watched repository. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubIssueCommentEvent(GithubEvent):
    """ A comment is added or updated on an issue. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubIssueEvent(GithubEvent):
    """ A new activity is registered on an issue. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubMemberEvent(GithubEvent):
    """ New activity related to repository collaborators. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubPublicEvent(GithubEvent):
    """ A private repository is made public. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubPullRequestEvent(GithubEvent):
    """ New activity related to a pull request. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubPullRequestReviewCommentEvent(GithubEvent):
    """ New activity related to comments of a pull request. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubReleaseEvent(GithubEvent):
    """ New activity related to the release of a repository. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubSponsorshipEvent(GithubEvent):
    """ New activity related to the sponsorship of a repository. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


class GithubWatchEvent(GithubEvent):
    """ Event triggered when someone stars or starts watching a repository. """

    def __init__(self, payload: dict, *args, **kwargs):
        super().__init__(*args, payload=payload, **kwargs)


# vim:sw=4:ts=4:et:



import os
import pathlib
import random
import shutil
import statistics
import time

from enum import Enum
from threading import Thread, Event
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Set,
    Union,
)
from platypush.config import Config
from platypush.context import get_bus
from platypush.entities import Entity, LightEntityManager
from platypush.entities.lights import Light as LightEntity
from platypush.message.event.light import (
    LightAnimationStartedEvent,
    LightAnimationStoppedEvent,
    LightStatusChangeEvent,
)
from platypush.plugins import RunnablePlugin, action


class LightHuePlugin(RunnablePlugin, LightEntityManager):
    """
    Philips Hue lights plugin.
    """

    MAX_BRI = 255
    MAX_SAT = 255
    MAX_HUE = 65535
    MIN_CT = 154
    MAX_CT = 500
    ANIMATION_CTRL_QUEUE_NAME = 'platypush/light/hue/AnimationCtrl'
    _BRIDGE_RECONNECT_SECONDS = 5
    _MAX_RECONNECT_TRIES = 5
    _UNINITIALIZED_BRIDGE_ERR = 'The Hue bridge is not initialized'

    class Animation(Enum):
        """
        Inner class to model light animations.
        """

        COLOR_TRANSITION = 'color_transition'
        BLINK = 'blink'

        def __eq__(self, other):
            """
            Check if the configuration of two light animations matches.
            """
            if isinstance(other, str):
                return self.value == other
            if isinstance(other, self.__class__):
                return self == other
            return False

    def __init__(
        self,
        bridge: str,
        lights: Optional[Iterable[str]] = None,
        groups: Optional[Iterable[str]] = None,
        poll_interval: float = 20.0,
        config_file: Optional[str] = None,
        **kwargs,
    ):
        """
        :param bridge: Bridge address or hostname
        :param lights: Default lights to be controlled (default: all)
        :param groups: Default groups to be controlled (default: all)
        :param poll_interval: How often the plugin should check the bridge for light
            updates (default: 20 seconds).
        :param config_file: Path to the phue configuration file containing the
            access token to authenticate to the Hue bridge and the bridge
            configuration (default: ``<WORKDIR>/light.hue/config.json``).
        """

        poll_seconds = kwargs.pop('poll_seconds', None)
        if poll_seconds is not None:
            self.logger.warning('poll_seconds is deprecated, use poll_interval instead')

            if poll_interval is None:
                poll_interval = poll_seconds

        super().__init__(**kwargs)

        self.bridge_address = bridge
        self.bridge = None
        self.logger.info(
            'Initializing Hue lights plugin - bridge: "%s"', self.bridge_address
        )

        self._init_config_file(config_file)
        self.connect()
        self.lights = set()
        self.groups = set()
        self.poll_interval = poll_interval
        self._cached_lights: Dict[str, dict] = {}

        if lights:
            self.lights = set(lights)
        elif groups:
            self.groups = set(groups)
            self.lights.update(self._expand_groups(self.groups))
        else:
            self.lights = {light['name'] for light in self._get_lights().values()}

        self.animation_thread = None
        self.animations: Dict[str, dict] = {}
        self._animation_stop = Event()
        self._init_animations()
        self.logger.info('Configured lights: %s', self.lights)

    def _init_config_file(self, config_file: Optional[str] = None):
        self.config_file = os.path.abspath(
            os.path.expanduser(
                config_file
                or os.path.join(Config.get_workdir(), 'light.hue', 'config.json')
            )
        )

        pathlib.Path(self.config_file).parent.mkdir(parents=True, exist_ok=True)

        # Check if the phue default ~/.python_hue file is present, and if so
        # copy it to our location
        legacy_config_file = os.path.join(os.path.expanduser('~'), '.python_hue')
        if os.path.isfile(legacy_config_file) and not os.path.exists(self.config_file):
            shutil.copy(legacy_config_file, self.config_file)

    def _expand_groups(self, groups: Iterable[str]) -> Set[str]:
        lights = set()
        light_id_to_name = {
            light_id: light['name'] for light_id, light in (self._get_lights().items())
        }

        groups_ = [g for g in self._get_groups().values() if g.get('name') in groups]

        for group in groups_:
            for light_id in group.get('lights', []):
                light_name = light_id_to_name.get(light_id)
                if light_name:
                    lights.add(light_name)

        return lights

    def _init_animations(self):
        self.animations = {
            'groups': {},
            'lights': {},
        }

        for group_id in self._get_groups():
            self.animations['groups'][group_id] = None
        for light_id in self._get_lights():
            self.animations['lights'][light_id] = None

    @action
    def connect(self):
        """
        Connect to the configured Hue bridge. If the device hasn't been paired
        yet, uncomment the ``.connect()`` and ``.get_api()`` lines and retry
        after clicking the pairing button on your bridge.
        """

        # Lazy init
        if not self.bridge:
            from phue import Bridge, PhueRegistrationException

            success = False
            n_tries = 0

            while not success:
                try:
                    n_tries += 1
                    self.bridge = Bridge(
                        self.bridge_address, config_file_path=self.config_file
                    )

                    success = True
                except PhueRegistrationException as e:
                    self.logger.warning('Bridge registration error: %s', e)

                    if n_tries >= self._MAX_RECONNECT_TRIES:
                        self.logger.error(
                            (
                                'Bridge registration failed after ' + '{} attempts'
                            ).format(n_tries)
                        )
                        break

                    time.sleep(self._BRIDGE_RECONNECT_SECONDS)

            self.logger.info('Bridge connected')
            self.get_scenes()
        else:
            self.logger.info('Bridge already connected')

    @action
    def get_scenes(self):
        """
        Get the available scenes on the devices.

        :returns: The scenes configured on the bridge.

        Example output::

            {
                "scene-id-1": {
                    "name": "Scene 1",
                    "lights": [
                        "1",
                        "3"
                    ],

                    "owner": "owner-id",
                    "recycle": true,
                    "locked": false,
                    "appdata": {},
                    "picture": "",
                    "lastupdated": "2018-06-01T00:00:00",
                    "version": 1
                }
            }

        """

        return {
            id: {
                'id': id,
                **scene,
            }
            for id, scene in self._get_scenes().items()
        }

    @action
    def get_lights(self):
        """
        Get the configured lights.

        :returns: List of available lights as id->dict.

        Example::

            {
                "1": {
                    "state": {
                        "on": true,
                        "bri": 254,
                        "hue": 1532,
                        "sat": 215,
                        "effect": "none",
                        "xy": [
                            0.6163,
                            0.3403
                        ],

                        "ct": 153,
                        "alert": "none",
                        "colormode": "hs",
                        "reachable": true
                    },

                    "type": "Extended color light",
                    "name": "Lightbulb 1",
                    "modelid": "LCT001",
                    "manufacturername": "Philips",
                    "uniqueid": "00:11:22:33:44:55:66:77-88",
                    "swversion": "5.105.0.21169"
                }
            }

        """

        return {
            id: {
                'id': id,
                **light,
            }
            for id, light in self._get_lights(publish_entities=True).items()
        }

    @action
    def get_groups(self):
        """
        Get the list of configured light groups.

        :returns: List of configured light groups as id->dict.

        Example::

            {
                "1": {
                    "name": "Living Room",
                    "lights": [
                        "16",
                        "13",
                        "12",
                        "11",
                        "10",
                        "9",
                        "1",
                        "3"
                    ],

                    "type": "Room",
                    "state": {
                        "all_on": true,
                        "any_on": true
                    },

                    "class": "Living room",
                    "action": {
                        "on": true,
                        "bri": 241,
                        "hue": 37947,
                        "sat": 221,
                        "effect": "none",
                        "xy": [
                            0.2844,
                            0.2609
                        ],

                        "ct": 153,
                        "alert": "none",
                        "colormode": "hs"
                    }
                }
            }

        """

        return {
            id: {
                'id': id,
                **group,
            }
            for id, group in self._get_groups().items()
        }

    @action
    def get_animations(self):
        """
        Get the list of running light animations.

        :returns: dict.

        Structure::

            {
                "groups": {
                    "id_1": {
                        "type": "color_transition",
                        "hue_range": [0,65535],
                        "sat_range": [0,255],
                        "bri_range": [0,255],
                        "hue_step": 10,
                        "sat_step": 10,
                        "bri_step": 2,
                        "transition_seconds": 2

                    }

                },

                "lights": {
                    "id_1": {}

                }

            }

        """

        return self.animations

    def _exec(self, attr, *args, **kwargs):
        try:
            self.connect()
            self.stop_animation()
        except Exception as e:
            # Reset bridge connection
            self.bridge = None
            raise e

        assert self.bridge, self._UNINITIALIZED_BRIDGE_ERR
        lights = []
        groups = []

        if 'lights' in kwargs:
            lights = (
                kwargs.pop('lights').split(',').strip()
                if isinstance(lights, str)
                else kwargs.pop('lights')
            )
        if 'groups' in kwargs:
            groups = (
                kwargs.pop('groups').split(',').strip()
                if isinstance(groups, str)
                else kwargs.pop('groups')
            )

        if not lights and not groups:
            lights = self.lights
            groups = self.groups

        if not self.bridge:
            self.connect()

        try:
            if attr == 'scene':
                assert groups, 'No groups specified'
                self.bridge.run_scene(list(groups)[0], kwargs.pop('name'))
            else:
                if groups:
                    self.bridge.set_group(list(groups), attr, *args, **kwargs)
                if lights:
                    self.bridge.set_light(list(lights), attr, *args, **kwargs)
        except Exception as e:
            # Reset bridge connection
            self.bridge = None
            raise e

        return self._get_lights(publish_entities=True)

    @action
    def set_lights(self, lights, *_, **kwargs):  # pylint: disable=arguments-differ
        """
        Set a set of properties on a set of lights.

        :param light: List of lights to set. Each item can represent a light
            name or ID.
        :param kwargs: key-value list of the parameters to set.

        Example call::

            {
                "type": "request",
                "action": "light.hue.set_lights",
                "args": {
                    "lights": ["Bulb 1", "Bulb 2"],
                    "sat": 255
                }
            }

        """

        self.connect()
        assert self.bridge, self._UNINITIALIZED_BRIDGE_ERR
        all_lights = self._get_lights()

        for i, l in enumerate(lights):
            if str(l) in all_lights:
                lights[i] = all_lights[str(l)]['name']

        # Convert entity attributes to local attributes
        if kwargs.get('saturation') is not None:
            kwargs['sat'] = kwargs.pop('saturation')
        if kwargs.get('brightness') is not None:
            kwargs['bri'] = kwargs.pop('brightness')
        if kwargs.get('temperature') is not None:
            kwargs['ct'] = kwargs.pop('temperature')

        # "Unroll" the map
        args = []
        for arg, value in kwargs.items():
            args += [arg, value]

        assert len(args) > 1, 'Not enough parameters passed to set_lights'
        param = args.pop(0)
        value = args.pop(0)
        self.bridge.set_light(lights, param, value, *args)
        return self._get_lights(publish_entities=True)

    @action
    def set_group(self, group, **kwargs):
        """
        Set a group (or groups) property.

        :param group: Group or groups to set. Can be a string representing the
            group name, a group object, a list of strings, or a list of group
            objects.
        :param kwargs: key-value list of parameters to set.

        Example call::

            {
                "type": "request",
                "action": "light.hue.set_group",
                "args": {
                    "light": "Living Room",
                    "sat": 255
                }
            }

        """

        self.connect()
        assert self.bridge, self._UNINITIALIZED_BRIDGE_ERR
        self.bridge.set_group(group, **kwargs)

    @action
    def on(  # pylint: disable=arguments-differ
        self, lights=None, groups=None, **kwargs
    ):
        """
        Turn lights/groups on.

        :param lights: Lights to turn on (names or light objects). Default: plugin default lights
        :param groups: Groups to turn on (names or group objects). Default: plugin default groups
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('on', True, lights=lights, groups=groups, **kwargs)

    @action
    def off(  # pylint: disable=arguments-differ
        self, lights=None, groups=None, **kwargs
    ):
        """
        Turn lights/groups off.

        :param lights: Lights to turn off (names or light objects). Default: plugin default lights
        :param groups: Groups to turn off (names or group objects). Default: plugin default groups
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('on', False, lights=lights, groups=groups, **kwargs)

    @action
    def toggle(  # pylint: disable=arguments-differ
        self, lights=None, groups=None, **kwargs
    ):
        """
        Toggle lights/groups on/off.

        :param lights: Lights to turn off (names or light objects). Default: plugin default lights
        :param groups: Groups to turn off (names or group objects). Default: plugin default groups
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        lights_on = []
        lights_off = []
        groups_on = []
        groups_off = []

        if groups:
            all_groups = self._get_groups().values()
            groups_on = [
                group['name']
                for group in all_groups
                if group['name'] in groups and group['state']['any_on'] is True
            ]

            groups_off = [
                group['name']
                for group in all_groups
                if group['name'] in groups and group['state']['any_on'] is False
            ]

        if not groups and not lights:
            lights = self.lights

        if lights:
            all_lights = self._get_lights()

            lights_on = [
                light['name']
                for light_id, light in all_lights.items()
                if (light_id in lights or light['name'] in lights)
                and light['state']['on'] is True
            ]

            lights_off = [
                light['name']
                for light_id, light in all_lights.items()
                if (light_id in lights or light['name'] in lights)
                and light['state']['on'] is False
            ]

        if lights_on or groups_on:
            self._exec('on', False, lights=lights_on, groups=groups_on, **kwargs)

        if lights_off or groups_off:
            self._exec('on', True, lights=lights_off, groups=groups_off, **kwargs)

    @action
    def bri(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups brightness.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param value: Brightness value (range: 0-255)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec(
            'bri',
            int(value) % (self.MAX_BRI + 1),
            lights=lights,
            groups=groups,
            **kwargs,
        )

    @action
    def sat(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups saturation.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param value: Saturation value (range: 0-255)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec(
            'sat',
            int(value) % (self.MAX_SAT + 1),
            lights=lights,
            groups=groups,
            **kwargs,
        )

    @action
    def hue(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups color hue.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param value: Hue value (range: 0-65535)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec(
            'hue',
            int(value) % (self.MAX_HUE + 1),
            lights=lights,
            groups=groups,
            **kwargs,
        )

    @action
    def xy(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups XY colors.

        :param value: xY value
        :type value: list[float] containing the two values
        :param lights: List of lights.
        :param groups: List of groups.
        """
        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('xy', value, lights=lights, groups=groups, **kwargs)

    @action
    def ct(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups color temperature.

        :param value: Temperature value (range: 154-500)
        :type value: int
        :param lights: List of lights.
        :param groups: List of groups.
        """
        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('ct', value, lights=lights, groups=groups, **kwargs)

    @action
    def delta_bri(self, delta, lights=None, groups=None, **kwargs):
        """
        Change lights/groups brightness by a delta [-100, 100] compared to the current state.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param delta: Brightness delta value (range: -100, 100)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []

        if lights:
            bri = statistics.mean(
                [
                    light['state']['bri']
                    for light in self._get_lights().values()
                    if light['name'] in lights
                ]
            )
        elif groups:
            bri = statistics.mean(
                [
                    group['action']['bri']
                    for group in self._get_groups().values()
                    if group['name'] in groups
                ]
            )
        else:
            bri = statistics.mean(
                [
                    light['state']['bri']
                    for light in self._get_lights().values()
                    if light['name'] in self.lights
                ]
            )

        delta *= self.MAX_BRI / 100
        if bri + delta < 0:
            bri = 0
        elif bri + delta > self.MAX_BRI:
            bri = self.MAX_BRI
        else:
            bri += delta

        return self._exec('bri', int(bri), lights=lights, groups=groups, **kwargs)

    @action
    def delta_sat(self, delta, lights=None, groups=None, **kwargs):
        """
        Change lights/groups saturation by a delta [-100, 100] compared to the current state.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param delta: Saturation delta value (range: -100, 100)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []

        if lights:
            sat = statistics.mean(
                [
                    light['state']['sat']
                    for light in self._get_lights().values()
                    if light['name'] in lights
                ]
            )
        elif groups:
            sat = statistics.mean(
                [
                    group['action']['sat']
                    for group in self._get_groups().values()
                    if group['name'] in groups
                ]
            )
        else:
            sat = statistics.mean(
                [
                    light['state']['sat']
                    for light in self._get_lights().values()
                    if light['name'] in self.lights
                ]
            )

        delta *= self.MAX_SAT / 100
        if sat + delta < 0:
            sat = 0
        elif sat + delta > self.MAX_SAT:
            sat = self.MAX_SAT
        else:
            sat += delta

        return self._exec('sat', int(sat), lights=lights, groups=groups, **kwargs)

    @action
    def delta_hue(self, delta, lights=None, groups=None, **kwargs):
        """
        Change lights/groups hue by a delta [-100, 100] compared to the current state.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param delta: Hue delta value (range: -100, 100)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []

        if lights:
            hue = statistics.mean(
                [
                    light['state']['hue']
                    for light in self._get_lights().values()
                    if light['name'] in lights
                ]
            )
        elif groups:
            hue = statistics.mean(
                [
                    group['action']['hue']
                    for group in self._get_groups().values()
                    if group['name'] in groups
                ]
            )
        else:
            hue = statistics.mean(
                [
                    light['state']['hue']
                    for light in self._get_lights().values()
                    if light['name'] in self.lights
                ]
            )

        delta *= self.MAX_HUE / 100
        if hue + delta < 0:
            hue = 0
        elif hue + delta > self.MAX_HUE:
            hue = self.MAX_HUE
        else:
            hue += delta

        return self._exec('hue', int(hue), lights=lights, groups=groups, **kwargs)

    @action
    def scene(self, name, lights=None, groups=None, **kwargs):
        """
        Set a scene by name.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param name: Name of the scene
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('scene', name=name, lights=lights, groups=groups, **kwargs)

    @action
    def is_animation_running(self):
        """
        :returns: True if there is an animation running, false otherwise.
        """

        return self.animation_thread is not None and self.animation_thread.is_alive()

    @action
    def stop_animation(self):
        """
        Stop a running animation.
        """
        if self.is_animation_running():
            self._animation_stop.set()
            self._init_animations()

    @action
    def animate(
        self,
        animation,
        duration=None,
        hue_range=None,
        sat_range=None,
        bri_range=None,
        lights=None,
        groups=None,
        hue_step=1000,
        sat_step=2,
        bri_step=1,
        transition_seconds=1.0,
    ):
        """
        Run a lights animation.

        :param animation: Animation name. Supported types: **color_transition** and **blink**
        :type animation: str

        :param duration: Animation duration in seconds (default: None, i.e. continue until stop)
        :type duration: float

        :param hue_range: If you selected a ``color_transition``, this will
            specify the hue range of your color ``color_transition``.
            Default: [0, 65535]
        :type hue_range: list[int]

        :param sat_range: If you selected a color ``color_transition``, this
            will specify the saturation range of your color
            ``color_transition``. Default: [0, 255]
        :type sat_range: list[int]

        :param bri_range: If you selected a color ``color_transition``, this
            will specify the brightness range of your color
            ``color_transition``. Default: [254, 255]
        :type bri_range: list[int]

        :param lights: Lights to control (names, IDs or light objects).
            Default: plugin default lights
        :param groups: Groups to control (names, IDs or group objects).
            Default: plugin default groups

        :param hue_step: If you selected a color ``color_transition``, this
            will specify by how much the color hue will change between
            iterations. Default: 1000
        :type hue_step: int

        :param sat_step: If you selected a color ``color_transition``, this
            will specify by how much the saturation will change
            between iterations. Default: 2
        :type sat_step: int

        :param bri_step: If you selected a color ``color_transition``, this
            will specify by how much the brightness will change between iterations.
            Default: 1
        :type bri_step: int

        :param transition_seconds: Time between two transitions or blinks in
            seconds. Default: 1.0
        :type transition_seconds: float
        """

        self.stop_animation()
        self._animation_stop.clear()
        all_lights = self._get_lights()
        bri_range = bri_range or [self.MAX_BRI - 1, self.MAX_BRI]
        sat_range = sat_range or [0, self.MAX_SAT]
        hue_range = hue_range or [0, self.MAX_HUE]

        if groups:
            groups = {
                group_id: group
                for group_id, group in self._get_groups().items()
                if group.get('name') in groups or group_id in groups
            }

            lights = set(lights or [])
            lights.update(self._expand_groups([g['name'] for g in groups.values()]))
        elif lights:
            lights = {
                light['name']
                for light_id, light in all_lights.items()
                if light['name'] in lights or int(light_id) in lights
            }
        else:
            lights = self.lights

        info = {
            'type': animation,
            'duration': duration,
            'hue_range': hue_range,
            'sat_range': sat_range,
            'bri_range': bri_range,
            'hue_step': hue_step,
            'sat_step': sat_step,
            'bri_step': bri_step,
            'transition_seconds': transition_seconds,
        }

        if groups:
            for group_id in groups:
                self.animations['groups'][group_id] = info

        for light_id, light in all_lights.items():
            if light['name'] in lights:
                self.animations['lights'][light_id] = info

        def _initialize_light_attrs(lights):
            lights_by_name = {
                light['name']: light for light in self._get_lights().values()
            }

            if animation == self.Animation.COLOR_TRANSITION:
                return {
                    light: {
                        **(
                            {'hue': random.randint(hue_range[0], hue_range[1])}  # type: ignore
                            if 'hue' in lights_by_name.get(light, {}).get('state', {})
                            else {}
                        ),
                        **(
                            {'sat': random.randint(sat_range[0], sat_range[1])}  # type: ignore
                            if 'sat' in lights_by_name.get(light, {}).get('state', {})
                            else {}
                        ),
                        **(
                            {'bri': random.randint(bri_range[0], bri_range[1])}  # type: ignore
                            if 'bri' in lights_by_name.get(light, {}).get('state', {})
                            else {}
                        ),
                    }
                    for light in lights
                }
            elif animation == self.Animation.BLINK:
                return {
                    light: {
                        'on': True,
                        **({'bri': self.MAX_BRI} if 'bri' in light else {}),
                        'transitiontime': 0,
                    }
                    for light in lights
                }

            raise AssertionError(f'Unknown animation type: {animation}')

        def _next_light_attrs(lights):
            if animation == self.Animation.COLOR_TRANSITION:
                for light, attrs in lights.items():
                    for attr, value in attrs.items():
                        if attr == 'hue':
                            attr_range = hue_range
                            attr_step = hue_step
                        elif attr == 'bri':
                            attr_range = bri_range
                            attr_step = bri_step
                        elif attr == 'sat':
                            attr_range = sat_range
                            attr_step = sat_step
                        else:
                            continue

                        lights[light][attr] = (
                            (value - attr_range[0] + attr_step)
                            % (attr_range[1] - attr_range[0] + 1)
                        ) + attr_range[0]
            elif animation == self.Animation.BLINK:
                lights = {
                    light: {
                        'on': not attrs['on'],
                        'bri': self.MAX_BRI,
                        'transitiontime': 0,
                    }
                    for (light, attrs) in lights.items()
                }

            return lights

        def _should_stop():
            return self._animation_stop.is_set()

        def _animate_thread(lights):
            get_bus().post(
                LightAnimationStartedEvent(
                    lights=lights,
                    groups=list((groups or {}).keys()),
                    animation=animation,
                )
            )

            lights = _initialize_light_attrs(lights)
            animation_start_time = time.time()
            stop_animation = False

            while not stop_animation and not (
                duration and time.time() - animation_start_time > duration
            ):
                assert self.bridge, self._UNINITIALIZED_BRIDGE_ERR

                try:
                    if animation == self.Animation.COLOR_TRANSITION:
                        for light, attrs in lights.items():
                            self.logger.debug('Setting %s to %s', lights, attrs)
                            self.bridge.set_light(light, attrs)
                    elif animation == self.Animation.BLINK:
                        conf = lights[list(lights.keys())[0]]
                        self.logger.debug('Setting lights to %s', conf)

                        if groups:
                            self.bridge.set_group(
                                [g['name'] for g in groups.values()], conf
                            )
                        else:
                            self.bridge.set_light(lights.keys(), conf)

                    if transition_seconds:
                        time.sleep(transition_seconds)

                    stop_animation = _should_stop()
                except Exception as e:
                    self.logger.warning(e)
                    time.sleep(2)

                lights = _next_light_attrs(lights)

            get_bus().post(
                LightAnimationStoppedEvent(
                    lights=list(lights.keys()),
                    groups=list((groups or {}).keys()),
                    animation=animation,
                )
            )

            self.animation_thread = None

        self.animation_thread = Thread(
            target=_animate_thread, name='HueAnimate', args=(lights,)
        )
        self.animation_thread.start()

    def _get_light_attr(self, light, attr: str):
        try:
            return getattr(light, attr, None)
        except KeyError:
            return None

    def transform_entities(
        self, entities: Union[Iterable[Union[dict, Entity]], Mapping[Any, dict]]
    ) -> Collection[Entity]:
        new_entities = []
        if isinstance(entities, dict):
            entities = [{'id': id, **e} for id, e in entities.items()]

        for entity in entities:
            if isinstance(entity, Entity):
                new_entities.append(entity)
            elif isinstance(entity, dict):
                new_entities.append(
                    LightEntity(
                        id=entity['id'],
                        name=entity['name'],
                        description=entity.get('type'),
                        on=entity.get('state', {}).get('on', False),
                        brightness=entity.get('state', {}).get('bri'),
                        saturation=entity.get('state', {}).get('sat'),
                        hue=entity.get('state', {}).get('hue'),
                        temperature=entity.get('state', {}).get('ct'),
                        colormode=entity.get('colormode'),
                        reachable=entity.get('state', {}).get('reachable'),
                        x=(
                            entity['state']['xy'][0]
                            if entity.get('state', {}).get('xy')
                            else None
                        ),
                        y=(
                            entity['state']['xy'][1]
                            if entity.get('state', {}).get('xy')
                            else None
                        ),
                        effect=entity.get('state', {}).get('effect'),
                        **(
                            {
                                'hue_min': 0,
                                'hue_max': self.MAX_HUE,
                            }
                            if entity.get('state', {}).get('hue') is not None
                            else {}
                        ),
                        **(
                            {
                                'saturation_min': 0,
                                'saturation_max': self.MAX_SAT,
                            }
                            if entity.get('state', {}).get('sat') is not None
                            else {}
                        ),
                        **(
                            {
                                'brightness_min': 0,
                                'brightness_max': self.MAX_BRI,
                            }
                            if entity.get('state', {}).get('bri') is not None
                            else {}
                        ),
                        **(
                            {
                                'temperature_min': self.MIN_CT,
                                'temperature_max': self.MAX_CT,
                            }
                            if entity.get('state', {}).get('ct') is not None
                            else {}
                        ),
                    )
                )

        return new_entities

    def _get_lights(self, publish_entities=False) -> dict:
        assert self.bridge, self._UNINITIALIZED_BRIDGE_ERR
        lights = self.bridge.get_light()
        lights = {id: light for id, light in lights.items() if not light.get('recycle')}
        self._cached_lights = lights
        if publish_entities:
            self.publish_entities(lights)  # type: ignore
        return lights

    def _get_groups(self) -> dict:
        assert self.bridge, self._UNINITIALIZED_BRIDGE_ERR
        groups = self.bridge.get_group() or {}
        return {id: group for id, group in groups.items() if not group.get('recycle')}

    def _get_scenes(self) -> dict:
        assert self.bridge, self._UNINITIALIZED_BRIDGE_ERR
        scenes = self.bridge.get_scene() or {}
        return {id: scene for id, scene in scenes.items() if not scene.get('recycle')}

    @action
    def status(self, *_, **__) -> Iterable[LightEntity]:
        lights = self.transform_entities(self._get_lights(publish_entities=True))
        for light in lights:
            light.id = light.external_id  # type: ignore
            for attr, value in (light.data or {}).items():
                setattr(light, attr, value)

            if light.external_id is not None:
                del light.external_id
            if light.data is not None:
                del light.data

        return lights

    def main(self):
        lights_prev = self._get_lights(publish_entities=True)  # Initialize the lights

        while not self.should_stop():
            try:
                lights_new = self._get_lights()
                for light_id, light in lights_new.items():
                    event_args = {}
                    new_state = light.get('state', {})
                    prev_state = lights_prev.get(light_id, {}).get('state', {})

                    for attr in ['on', 'bri', 'sat', 'hue', 'ct', 'xy']:
                        if attr in new_state and new_state.get(attr) != prev_state.get(
                            attr
                        ):
                            event_args[attr] = new_state.get(attr)

                    if event_args:
                        event_args['plugin_name'] = 'light.hue'
                        event_args['light_id'] = light_id
                        event_args['light_name'] = light.get('name')
                        get_bus().post(LightStatusChangeEvent(**event_args))
                        self.publish_entities([{'id': light_id, **light}])  # type: ignore

                lights_prev = lights_new
            finally:
                self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:
